"""
Analysis of responses to moving edges or bars.

Info:
    Relies on xarray dataset format defined in `flyvis.analysis.stimulus_responses`.
"""

from __future__ import annotations

from typing import Iterable, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import xarray as xr
from scipy.interpolate import interp1d

from flyvis.utils import groundtruth_utils, nodes_edges_utils
from flyvis.utils.color_utils import OFF, ON

from .visualization import plt_utils
from .visualization.plots import polar, violin_groups

__all__ = [
    "peak_responses",
    "peak_responses_angular",
    "direction_selectivity_index",
    "dsi_correlation_to_known",
    "dsi_violins_on_and_off",
    "dsi_violins",
    "preferred_direction",
    "plot_angular_tuning",
    "plot_T4_tuning",
    "plot_T5_tuning",
    "get_known_tuning_curves",
    "correlation_to_known_tuning_curves",
    "angular_distance_to_known",
]


def peak_responses(
    dataset: xr.Dataset,
    norm: xr.DataArray = None,
    from_degree: float = None,
    to_degree: float = None,
) -> xr.DataArray:
    """
    Compute peak responses from rectified voltages, optionally normalized.

    Args:
        dataset: Input dataset containing 'responses' and necessary coordinates.
        norm: Normalization array.
        from_degree: Starting degree for masking.
        to_degree: Ending degree for masking.

    Returns:
        Peak responses with reshaped and transposed dimensions.
    """
    config = dataset.attrs['config']
    from_degree = from_degree if from_degree is not None else config['offsets'][0] * 2.25
    to_degree = to_degree if to_degree is not None else (config['offsets'][1] - 1) * 2.25

    # Generate time masks
    masks = get_time_masks(
        dataset, from_column=from_degree / 5.8, to_column=to_degree / 5.8
    )

    # Apply masks to responses and rectify
    responses = dataset['responses']
    masked = responses.where(masks, other=0)
    rectified = masked.clip(min=0)  # Rectify: max(0, response)

    # Normalize if provided
    if norm is not None:
        rectified = rectified / norm

    # Compute peak (maximum over 'frame')
    peak = rectified.max(dim='frame')
    return peak


def get_time_masks(
    dataset: xr.Dataset, from_column: float = -1.5, to_column: float = 1.5
) -> xr.DataArray:
    """
    Generate time masks for each sample based on speed and column range.

    Args:
        dataset: Input dataset containing 'speed' and 'time' coordinates.
        from_column: Start of the column range.
        to_column: End of the column range.

    Returns:
        Boolean mask with dimensions ('sample', 'frame').
    """
    speeds = dataset['speed'].values
    unique_speeds = np.unique(speeds)
    config = dataset.attrs['config']
    start, end = config['offsets']
    times = dataset['time'].values

    # Precompute masks for unique speeds
    mask_dict = {}
    for speed in unique_speeds:
        t_start, t_end = time_window(
            speed, from_column=from_column, to_column=to_column, start=start, end=end
        )
        mask_dict[speed] = mask_between_seconds(t_start, t_end, times)

    # Map masks to each sample based on its speed
    masks = np.array([mask_dict[speed] for speed in speeds])

    # Create a DataArray for the masks
    mask_da = xr.DataArray(
        data=masks,
        dims=('sample', 'frame'),
        coords={'sample': dataset['sample'], 'frame': dataset['frame']},
    )

    return mask_da


def peak_responses_angular(
    dataset: xr.Dataset,
    norm: xr.DataArray = None,
    from_degree: float = None,
    to_degree: float = None,
) -> xr.DataArray:
    """
    Compute peak responses and make them complex over angles.

    Args:
        dataset: Input dataset.
        norm: Normalization array.
        from_degree: Starting degree for masking.
        to_degree: Ending degree for masking.

    Returns:
        Complex-valued peak responses.
    """
    peak = peak_responses(
        dataset, norm=norm, from_degree=from_degree, to_degree=to_degree
    )

    # Make complex over angles
    angles = peak['angle'].values
    radians = np.deg2rad(angles)
    # Expand dimensions to match broadcasting
    radians = radians[np.newaxis, :, np.newaxis]
    complex_peak = peak * np.exp(1j * radians)

    return complex_peak


def direction_selectivity_index(
    dataset: xr.Dataset,
    average: bool = True,
    norm: xr.DataArray = None,
    from_degree: float = None,
    to_degree: float = None,
) -> xr.DataArray:
    """
    Compute Direction Selectivity Index (DSI).

    Args:
        dataset: Input dataset.
        average: Whether to average over 'width' and 'speed'.
        norm: Normalization array.
        from_degree: Starting degree for masking.
        to_degree: Ending degree for masking.

    Returns:
        Direction Selectivity Index.
    """
    view = peak_responses_angular(
        dataset, norm=norm, from_degree=from_degree, to_degree=to_degree
    )
    view = view.set_index(sample=["angle", "width", "intensity", "speed"]).unstack(
        "sample"
    )

    # Compute vector sum over 'angle'
    vector_sum = view.sum(dim='angle')
    vector_length = np.abs(vector_sum)

    # Normalization: sum of absolute responses
    normalization = np.abs(view).sum(dim='angle').max(dim='intensity')
    dsi = vector_length / (normalization + 1e-15)

    if average:
        # Average over 'width' and 'speed'
        dsi = dsi.mean(dim=['width', 'speed'])

    return dsi.squeeze()


# -- plot ------------


def prepare_dsi_data(
    dsis, cell_types, sorted_type_list, known_on_off_first, sort_descending
):
    """
    Prepare DSI data for plotting.

    Args:
        dsis: Array of DSI values.
        cell_types: Array of cell type labels.
        sorted_type_list: List of cell types in desired order.
        known_on_off_first: Whether to sort known ON/OFF types first.
        sort_descending: Whether to sort DSIs in descending order.

    Returns:
        Tuple of prepared DSIs and cell types.
    """
    if known_on_off_first:
        sorted_type_list = nodes_edges_utils.nodes_list_sorting_on_off_unknown(cell_types)

    if sorted_type_list is not None:
        dsis = nodes_edges_utils.sort_by_mapping_lists(
            cell_types, sorted_type_list, dsis, axis=0
        )
        cell_types = np.array(sorted_type_list)

    if sort_descending:
        medians = np.median(dsis, axis=(-2, -1))
        index = np.argsort(medians)[::-1]
        dsis = dsis[index]
        cell_types = cell_types[index]

    return dsis, cell_types


def dsi_violins(
    dsis,
    cell_types,
    scatter_best=False,
    scatter_all=True,
    cmap=None,
    colors=None,
    color="b",
    figsize=[10, 1],
    fontsize=6,
    showmeans=False,
    showmedians=True,
    sorted_type_list=None,
    sort_descending=False,
    known_on_off_first=True,
    scatter_kwargs={},
    **kwargs,
):
    """
    Create violin plots for Direction Selectivity Index (DSI) data.

    Args:
        dsis: Array of DSI values.
        cell_types: Array of cell type labels.
        scatter_best: Whether to scatter the best points.
        scatter_all: Whether to scatter all points.
        cmap: Colormap for the violins.
        colors: Specific colors for the violins.
        color: Default color if colors is None and cmap is None.
        figsize: Figure size.
        fontsize: Font size for labels.
        showmeans: Whether to show means on violins.
        showmedians: Whether to show medians on violins.
        sorted_type_list: List of cell types in desired order.
        sort_descending: Whether to sort DSIs in descending order.
        known_on_off_first: Whether to sort known ON/OFF types first.
        **kwargs: Additional keyword arguments for violin_groups.

    Returns:
        Tuple of (figure, axis, colors, prepared DSIs)
    """
    dsis, cell_types = prepare_dsi_data(
        dsis, cell_types, sorted_type_list, known_on_off_first, sort_descending
    )

    if len(dsis.shape) == 1:
        dsis = dsis[None, None, :]
    elif len(dsis.shape) == 2:
        dsis = dsis[:, None]

    if colors is None and cmap is None and color is not None:
        colors = (color,)

    fig, ax, colors = violin_groups(
        dsis,
        cell_types[:],
        rotation=90,
        scatter=False,
        cmap=cmap,
        colors=colors,
        fontsize=fontsize,
        figsize=figsize,
        width=0.7,
        showmeans=showmeans,
        showmedians=showmedians,
        **kwargs,
    )

    if dsis.shape[1] == 1:
        plt_utils.scatter_on_violins_with_best(
            dsis.T.squeeze(), ax, scatter_best, scatter_all, **scatter_kwargs
        )

    return fig, ax, colors, dsis


def dsi_violins_on_and_off(
    dsis: xr.DataArray,
    cell_types: xr.DataArray,
    scatter_best=False,
    scatter_all=True,
    bold_output_type_labels=False,
    output_cell_types=None,
    known_on_off_first=True,
    sorted_type_list=None,
    figsize=[10, 1],
    ylim=(0, 1),
    color_known_types=True,
    fontsize=6,
    fig=None,
    axes=None,
    **kwargs,
):
    """
    Plot Direction Selectivity Index (DSI) for ON and OFF intensities.

    Args:
        dsis: DataArray of DSI values.
        cell_types: DataArray of cell type labels.
        scatter_best: Whether to scatter the best points.
        scatter_all: Whether to scatter all points.
        bold_output_type_labels: Whether to bold output type labels.
        output_cell_types: Cell types to output.
        known_on_off_first: Whether to sort known ON/OFF types first.
        sorted_type_list: List of cell types in desired order.
        figsize: Figure size.
        ylim: Y-axis limits.
        color_known_types: Whether to color known cell types.
        fontsize: Font size for labels.
        fig: Existing figure to use.
        axes: Existing axes to use.
        **kwargs: Additional keyword arguments for dsi_violins.

    Returns:
        Tuple of (figure, (ax1, ax2))
    """
    if len(dsis.shape) == 2:
        dsis = dsis[None, :]

    if fig is None or axes is None:
        fig, axes = plt.subplots(2, 1, figsize=figsize, sharex=True)
        plt_utils.rm_spines(axes[0], spines=("bottom",))

    for i, intensity in enumerate([1, 0]):
        color = ON if intensity == 1 else OFF
        _, ax, *_ = dsi_violins(
            dsis=dsis.sel(intensity=intensity).values.T,
            cell_types=cell_types.values,
            color=color,
            fig=fig,
            ax=axes[i],
            fontsize=fontsize,
            sorted_type_list=sorted_type_list,
            scatter_best=scatter_best,
            scatter_all=scatter_all,
            known_on_off_first=known_on_off_first,
            **kwargs,
        )

        ax.grid(False)
        ax.set_ylim(*ylim)
        plt_utils.trim_axis(ax)
        plt_utils.set_spine_tick_params(
            ax, tickwidth=0.5, ticklength=3, ticklabelpad=2, spinewidth=0.5
        )

        if bold_output_type_labels:
            plt_utils.boldify_labels(output_cell_types, ax)

        if color_known_types:
            plt_utils.color_labels(["T4a", "T4b", "T4c", "T4d"], ON, ax)
            plt_utils.color_labels(["T5a", "T5b", "T5c", "T5d"], OFF, ax)

    # axes[0].set_xticks([])
    axes[0].set_yticks(np.arange(0, 1.2, 0.5))
    axes[1].set_yticks(np.arange(0, 1.2, 0.5))
    axes[1].invert_yaxis()

    return fig, axes


def dsi_correlation_to_known(
    dsis: xr.DataArray, max_aggregate_dims=("intensity",)
) -> xr.DataArray:
    """
    Compute the correlation between predicted DSIs and known DSIs.

    Args:
        dsis: DataArray containing DSIs for ON and OFF intensities.
            Should have dimensions including 'intensity' and 'neuron',
            and a coordinate 'cell_type'.
        max_aggregate_dims: Dimensions to max-aggregate before computing correlation.

    Returns:
        Correlation between predicted and known DSIs.

    Note:
        Known DSIs are binary (0 or 1) based on whether the cell type
        is known to be motion-tuned.
    """
    # Ensure the 'intensity' dimension has length 2
    assert dsis.sizes['intensity'] == 2, "Dimension 'intensity' should have length 2"

    # Retrieve ground truth motion tuning information
    motion_tuning = groundtruth_utils.motion_tuning
    known_dsi_types = groundtruth_utils.known_dsi_types

    # Select dsis for known cell types
    dsis_for_known = dsis.where(dsis['cell_type'].isin(known_dsi_types), drop=True).max(
        dim=max_aggregate_dims
    )

    # Construct ground truth motion tuning array
    groundtruth_mt = xr.DataArray(
        [
            1.0 if ct in motion_tuning else 0.0
            for ct in dsis_for_known['cell_type'].values
        ],
        coords={'neuron': dsis_for_known['neuron']},
        dims=['neuron'],
    )

    # Compute correlation along 'neuron' dimension
    corr_dsi = xr.corr(dsis_for_known, groundtruth_mt, dim='neuron')

    return corr_dsi


def correlation_to_known_tuning_curves(
    dataset: xr.Dataset, absmax: bool = False
) -> xr.DataArray:
    """
    Compute correlation between predicted and known tuning curves.

    Args:
        dataset: Input dataset.
        absmax: If True, maximize magnitude of correlation regardless of sign.

    Returns:
        Correlation values for each cell type.
    """
    tuning = peak_responses(dataset)
    gt_tuning = get_known_tuning_curves(
        ["T4a", "T4b", "T4c", "T4d", "T5a", "T5b", "T5c", "T5d"], np.arange(0, 360, 30)
    )

    tuning = (
        tuning.set_index(sample=["angle", "intensity", "width", "speed"])
        .unstack("sample")
        .fillna(0.0)
        .custom.where(cell_type=["T4a", "T4b", "T4c", "T4d", "T5a", "T5b", "T5c", "T5d"])
    )

    # reset the neuron axis to make it compatible with the ground truth tuning curves
    tuning["neuron"] = np.arange(tuning.coords["neuron"].size)

    correlation = xr.corr(tuning, gt_tuning, dim="angle")
    correlation = correlation.fillna(0.0)
    if absmax:
        # take speed and width that maximize the magnitude of the correlation, regardless
        # of the sign
        argmax = np.abs(correlation).argmax(dim=("speed", "width"))
    else:
        # take speed and width that maximize the correlation as an experimentalist
        # would do
        argmax = correlation.argmax(dim=("speed", "width"))
    correlation = correlation.isel(argmax)
    return correlation


def get_known_tuning_curves(cell_types: List[str], angles: np.ndarray) -> xr.DataArray:
    """
    Retrieve ground truth tuning curves for specified cell types.

    Args:
        cell_types: List of cell type names.
        angles: Array of angles to interpolate curves to.

    Returns:
        DataArray of interpolated ground truth tuning curves.
    """
    gt_angles = np.arange(0, 360, 30)
    tuning_curves = []

    for cell_type in cell_types:
        gt_tuning = groundtruth_utils.tuning_curves[cell_type]
        interp_func = interp1d(
            gt_angles, gt_tuning, kind='cubic', fill_value="extrapolate"
        )
        gt_tuning = interp_func(angles)
        tuning_curves.append(gt_tuning)

    dataset = xr.DataArray(
        np.array(tuning_curves),
        dims=['neuron', 'angle'],
        coords={'cell_type': ("neuron", cell_types), 'angle': angles},
    )

    return dataset


def preferred_direction(
    dataset: xr.Dataset,
    average: bool = True,
    norm: xr.DataArray = None,
    from_degree: float = None,
    to_degree: float = None,
) -> xr.DataArray:
    """
    Compute the preferred direction based on peak responses.

    Args:
        dataset: Input dataset.
        average: Whether to average over certain dimensions.
        norm: Normalization array.
        from_degree: Starting degree for masking.
        to_degree: Ending degree for masking.

    Returns:
        Preferred direction angles in radians.
    """
    view = peak_responses_angular(
        dataset, norm=norm, from_degree=from_degree, to_degree=to_degree
    )
    view = view.set_index(sample=["angle", "width", "intensity", "speed"]).unstack(
        "sample"
    )

    # Compute vector sum over 'angle'
    vector_sum = view.sum(dim='angle')
    theta_pref = np.angle(vector_sum)

    if average:
        # Sum over 'width' and 'speed' before computing angle
        vector_sum = view.sum(dim=['width', 'speed', 'angle'])
        theta_pref = np.angle(vector_sum)

    vector_sum.data = theta_pref
    return vector_sum


def angular_distance_to_known(pds: xr.DataArray) -> xr.DataArray:
    """
    Compute angular distance between predicted and known preferred directions for T4/T5.

    Args:
        pds: Preferred directions for cells.

    Returns:
        Angular distances to known preferred directions.
    """
    t4s = pds.custom.where(cell_type=["T4a", "T4b", "T4c", "T4d"], intensity=1)
    t4_distances = angular_distances(t4s, np.array([np.pi, 0, np.pi / 2, 3 * np.pi / 2]))
    t5s = pds.custom.where(cell_type=["T5a", "T5b", "T5c", "T5d"], intensity=0)
    t5_distances = angular_distances(t5s, np.array([np.pi, 0, np.pi / 2, 3 * np.pi / 2]))
    # concatenate both xarrays again in the neuron dimension, drop intensity
    return xr.concat(
        [t4_distances.drop('intensity'), t5_distances.drop('intensity')], dim='neuron'
    )


def angular_distances(x: xr.DataArray, y: np.array, upper: float = np.pi) -> xr.DataArray:
    """
    Compute angular distances between two sets of angles.

    Args:
        x: First set of angles.
        y: Second set of angles.
        upper: Upper bound for distance calculation.

    Returns:
        Angular distances.
    """
    assert x.neuron.size == len(y)
    y_da = xr.DataArray(y, dims=['neuron'], coords={'neuron': x.coords['neuron']})

    result = xr.apply_ufunc(
        simple_angle_distance,
        x,
        y_da,
        input_core_dims=[['neuron'], ['neuron']],
        output_core_dims=[['neuron']],
        vectorize=True,
        kwargs={'upper': upper},
    )

    return result


def simple_angle_distance(
    a: np.ndarray, b: np.ndarray, upper: float = np.pi
) -> np.ndarray:
    """
    Calculate element-wise angle distance between 0 and pi radians.

    Args:
        a: First set of angles in radians.
        b: Second set of angles in radians.
        upper: Upper bound for distance calculation.

    Returns:
        Distance between 0 and pi radians.
    """
    a = np.atleast_1d(a)
    b = np.atleast_1d(b)

    # make all angles positive between 0 and 2 * pi
    a = a % (2 * np.pi)
    b = b % (2 * np.pi)

    y = np.zeros_like(a)
    # subtract the smaller angle from the larger one
    mask = a >= b
    y[mask] = a[mask] - b[mask]
    y[~mask] = b[~mask] - a[~mask]

    # map distances between pi and 2 pi to 0 and pi
    y[y > np.pi] = 2 * np.pi - y[y > np.pi]

    # map distances between 0 and pi to 0 and upper
    return y / np.pi * upper


def plot_angular_tuning(
    dataset: xr.Dataset,
    cell_type: int,
    intensity: int,
    figsize: Tuple[float, float] = (1, 1),
    fontsize: int = 5,
    linewidth: float = 1,
    anglepad: float = -7,
    xlabelpad: float = -1,
    groundtruth: bool = True,
    groundtruth_linewidth: float = 1.0,
    fig: plt.Figure = None,
    ax: plt.Axes = None,
    peak_responses_da: xr.DataArray = None,
    weighted_average: xr.DataArray = None,
    average_models: bool = False,
    colors: str = None,
    zorder: Union[int, Iterable] = 0,
    **kwargs,
) -> Tuple[plt.Figure, plt.Axes]:
    """
    Plot angular tuning for a specific cell type and intensity.

    Args:
        dataset: Input dataset.
        cell_type: Neuron index to plot.
        intensity: Intensity level (0 or 1).
        figsize: Figure size.
        fontsize: Font size.
        linewidth: Line width.
        anglepad: Angle padding.
        xlabelpad: X-label padding.
        groundtruth: Whether to plot ground truth.
        groundtruth_linewidth: Line width for ground truth.
        fig: Existing figure.
        ax: Existing axes.
        peak_responses_da: Precomputed peak responses.
        weighted_average: Weights for averaging models.
        average_models: Whether to average across models.
        colors: Color for the plot.
        zorder: Z-order for plotting.
        **kwargs: Additional keyword arguments for plotting.

    Returns:
        The figure and axes objects.
    """
    if peak_responses_da is None:
        peak_responses_da = peak_responses(dataset)

    peak_responses_da = peak_responses_da.set_index(
        sample=["angle", "width", "intensity", "speed"]
    ).unstack("sample")

    # Select the specific cell type
    peak = peak_responses_da.custom.where(cell_type=cell_type, intensity=intensity)

    # Squeeze irrelevant dimensions
    # peak = peak.squeeze(dim=['width', 'intensity', 'speed'], drop=True)

    # Average over speeds
    average_tuning = peak.mean(dim=('speed', 'width'))

    # Average over models if specified
    if average_models and weighted_average is not None:
        average_tuning = average_tuning.weighted(weighted_average).mean(dim='network_id')
    elif average_models:
        average_tuning = average_tuning.mean(dim='network_id')

    color = (ON if intensity == 1 else OFF) if colors is None else colors

    average_tuning = average_tuning / average_tuning.max()

    angles = average_tuning['angle'].values
    fig, ax = polar(
        angles,
        average_tuning.data.squeeze().T,
        figsize=figsize,
        fontsize=fontsize,
        linewidth=linewidth,
        anglepad=anglepad,
        xlabelpad=xlabelpad,
        color=color,
        fig=fig,
        ax=ax,
        zorder=zorder,
        **kwargs,
    )

    if groundtruth and cell_type in groundtruth_utils.tuning_curves:
        r_gt = np.array(groundtruth_utils.tuning_curves[cell_type])
        r_gt = r_gt / np.max(np.abs(r_gt))
        theta_gt = np.arange(0, 360, 360 / len(r_gt))
        polar(
            theta_gt,
            r_gt,
            figsize=figsize,
            fontsize=fontsize,
            linewidth=groundtruth_linewidth,
            anglepad=anglepad,
            xlabelpad=xlabelpad,
            color="k",
            fig=fig,
            ax=ax,
            # **kwargs,
        )

    return fig, ax


def plot_T4_tuning(dataset: xr.Dataset) -> None:
    """
    Plot tuning curves for T4 cells.

    Args:
        dataset: Input dataset.
    """
    fig, axes, _ = plt_utils.get_axis_grid(
        range(4),
        projection="polar",
        aspect_ratio=4,
        figsize=[2.95, 0.83],
        wspace=0.25,
    )
    for i, cell_type in enumerate(["T4a", "T4b", "T4c", "T4d"]):
        plot_angular_tuning(
            dataset,
            cell_type,
            intensity=1,
            fig=fig,
            ax=axes[i],
            groundtruth=True,
            aggregate_models="mean",
            linewidth=1.0,
        )
        axes[i].set_xlabel(cell_type)

    for ax in axes:
        ax.xaxis.label.set_fontsize(5)
        [i.set_linewidth(0.5) for i in ax.spines.values()]
        ax.grid(True, linewidth=0.5)


def plot_T5_tuning(dataset: xr.Dataset) -> None:
    """
    Plot tuning curves for T5 cells.

    Args:
        dataset: Input dataset.
    """
    fig, axes, _ = plt_utils.get_axis_grid(
        range(4),
        projection="polar",
        aspect_ratio=4,
        figsize=[2.95, 0.83],
        wspace=0.25,
    )
    for i, cell_type in enumerate(["T5a", "T5b", "T5c", "T5d"]):
        plot_angular_tuning(
            dataset,
            cell_type,
            intensity=0,
            fig=fig,
            ax=axes[i],
            groundtruth=True,
            aggregate_models="mean",
            linewidth=1.0,
        )
        axes[i].set_xlabel(cell_type)

    for ax in axes:
        ax.xaxis.label.set_fontsize(5)
        [i.set_linewidth(0.5) for i in ax.spines.values()]
        ax.grid(True, linewidth=0.5)


def mask_between_seconds(
    t_start: float,
    t_end: float,
    time: np.ndarray = None,
    t_pre: float = None,
    t_stim: float = None,
    t_post: float = None,
    dt: float = None,
) -> np.ndarray:
    """
    Create a boolean mask for time values between t_start and t_end.

    Args:
        t_start: Start time for the mask.
        t_end: End time for the mask.
        time: Array of time values. If None, it will be generated using other parameters.
        t_pre: Time before stimulus onset.
        t_stim: Stimulus duration.
        t_post: Time after stimulus offset.
        dt: Time step.

    Returns:
        Boolean mask array.

    Note:
        If 'time' is not provided, it will be generated using t_pre, t_stim, t_post,
        and dt.
    """
    time = time if time is not None else np.arange(-t_pre, t_stim + t_post - dt, dt)
    return (time >= t_start) & (time <= t_end)


def time_window(
    speed: float,
    from_column: float = -1.5,
    to_column: float = 1.5,
    start: float = -10,
    end: float = 11,
) -> tuple[float, float]:
    """
    Calculate start and end time when the bar passes from_column to to_column.

    Args:
        speed: Speed in columns/s (5.8deg/s).
        from_column: Starting column in 5.8deg units.
        to_column: Ending column in 5.8deg units.
        start: Starting position in LED units (2.25deg).
        end: Ending position in LED units (2.25deg).

    Returns:
        Tuple containing start and end times.

    Note:
        The function adjusts the to_column by adding a single LED width (2.25 deg)
        to make it symmetric around the central column.
    """
    start_in_columns = start * 2.25 / 5.8  # in 5.8deg
    end_in_columns = end * 2.25 / 5.8  # in 5.8deg

    # Make it symmetric around the central column by adding a single LED width
    to_column += 2.25 / 5.8

    assert abs(start_in_columns) >= abs(from_column)
    assert abs(end_in_columns) >= abs(to_column)

    # Calculate when the edge is at the from_column
    t_start = (abs(start_in_columns) - abs(from_column)) / speed
    # Calculate when it's at the to_column
    t_end = t_start + (to_column - from_column) / speed
    return t_start, t_end



import logging
import os
import tempfile
from pathlib import Path
from time import sleep
from typing import Any, Iterable, Literal, Optional, Union

import ffmpeg
import matplotlib
from IPython import display

__all__ = ["Animation", "AnimationCollector", "convert"]

TESTING = os.getenv("TESTING", "False").lower() == "true"
COLAB = bool(os.getenv("COLAB_RELEASE_TAG"))
ANIMATION_DIR = os.getenv("ANIMATION_DIR", "animations")


class Animation:
    """Base class for animations.

    Subclasses must implement `init` and `animate` methods.

    Args:
        path: Path to save the animation.
        fig: Existing Figure instance or None.
        suffix: Suffix for the animation path.

    Attributes:
        fig (matplotlib.figure.Figure): Figure instance for the animation.
        update (bool): Whether to update the canvas after each animation step.
        batch_sample (int): Sample to animate.
        frames (int): Number of frames in the animation.
        n_samples (int): Number of samples in the animation.
        path (Path): Path to save the animation.
    """

    fig: Optional[matplotlib.figure.Figure] = None
    update: bool = True
    batch_sample: int = 0
    frames: int = 0
    n_samples: int = 0

    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        fig: Optional[matplotlib.figure.Figure] = None,
        suffix: str = "{}",
    ):
        self.path = Path(ANIMATION_DIR if path is None else path) / suffix.format(
            self.__class__.__name__
        )
        self.fig = fig

    def init(self, frame: int = 0) -> None:
        """Initialize the animation.

        Args:
            frame: Initial frame number.

        Raises:
            NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError("Subclasses should implement this method.")

    def animate(self, frame: int) -> None:
        """Animate a single frame.

        Args:
            frame: Frame number to animate.

        Raises:
            NotImplementedError: If not implemented by subclass.
        """
        raise NotImplementedError("Subclasses should implement this method.")

    def update_figure(self, clear_output: bool = True) -> None:
        """Update the figure canvas.

        Args:
            clear_output: Whether to clear the previous output.
        """
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()
        if matplotlib.get_backend().lower() != "nbagg" or COLAB:
            display.display(self.fig)
            if clear_output:
                display.clear_output(wait=True)

    def animate_save(self, frame: int, dpi: int = 100) -> None:
        """Update the figure to the given frame and save it.

        Args:
            frame: Frame number to animate and save.
            dpi: Dots per inch for the saved image.
        """
        self.animate(frame)
        identifier = f"{self.batch_sample:04}_{frame:04}"
        self.fig.savefig(
            self._path / f"{identifier}.png",
            dpi=dpi,
            bbox_inches="tight",
            facecolor=self.fig.get_facecolor(),
            edgecolor="none",
        )

    def _get_indices(self, key: str, input: Union[str, Iterable]) -> list[int]:
        """Get sorted list of indices based on input.

        Args:
            key: Attribute name to get total number of indices.
            input: Input specifying which indices to return.

        Returns:
            Sorted list of indices.

        Raises:
            ValueError: If input is invalid.
        """
        total = getattr(self, key)
        _indices = set(range(total))
        if isinstance(input, str) and input == "all":
            indices = _indices
        elif isinstance(input, Iterable):
            indices = _indices.intersection(set(input))
        else:
            raise ValueError(f"Invalid input for {key}: {input}")
        return sorted(indices)

    def animate_in_notebook(
        self,
        frames: Union[str, Iterable] = "all",
        samples: Union[str, Iterable] = "all",
        repeat: int = 1,
    ) -> None:
        """Play animation within a Jupyter notebook.

        Args:
            frames: Frames to animate.
            samples: Samples to animate.
            repeat: Number of times to repeat the animation.
        """
        frames_list, samples_list = self._initialize_animation(frames, samples)

        if TESTING:
            frames_list = frames_list[:1]
            samples_list = samples_list[:1]
            repeat = 1

        try:
            for _ in range(repeat):
                for sample in samples_list:
                    self.batch_sample = sample
                    for frame in frames_list:
                        self.animate(frame)
                        sleep(0.1)  # Pause between frames
        except KeyboardInterrupt:
            print("Animation interrupted. Displaying last frame.")
            self.update_figure(clear_output=False)
            return

    def _verify_backend(self) -> None:
        """Ensure the notebook backend is set correctly.

        Raises:
            RuntimeError: If matplotlib backend is not set to notebook.
        """
        backend = matplotlib.get_backend().lower()
        if backend != "nbagg" and not COLAB:
            raise RuntimeError(
                "Matplotlib backend is not set to notebook. Use '%matplotlib notebook'."
            )

    def _initialize_animation(
        self, frames: Union[str, Iterable], samples: Union[str, Iterable]
    ) -> tuple[list[int], list[int]]:
        """Initialize the animation state.

        Args:
            frames: Frames to animate.
            samples: Samples to animate.

        Returns:
            Tuple of frames list and samples list.
        """
        self.update = True
        self.init()
        frames_list = self._get_indices("frames", frames)
        samples_list = self._get_indices("n_samples", samples)
        return frames_list, samples_list

    def plot(self, sample: int, frame: int) -> None:
        """Plot a single frame for a specific sample.

        Args:
            sample: Sample number to plot.
            frame: Frame number to plot.
        """
        previous_sample = self.batch_sample
        self.update = True
        self.init()
        self.batch_sample = sample
        self.animate(frame)
        self.batch_sample = previous_sample

    def _create_temp_dir(self, path: Optional[Union[str, Path]] = None) -> None:
        """Create a temporary directory as destination for the images.

        Args:
            path: Path to create the temporary directory.
        """
        self._temp_dir = tempfile.TemporaryDirectory()
        self._path = Path(self._temp_dir.name)

    def to_vid(
        self,
        fname: str,
        frames: Union[str, Iterable] = "all",
        dpi: int = 100,
        framerate: int = 30,
        samples: Union[str, Iterable] = "all",
        delete_if_exists: bool = False,
        source_path: Optional[Union[str, Path]] = None,
        dest_path: Optional[Union[str, Path]] = None,
        type: Literal["mp4", "webm"] = "webm",
    ) -> None:
        """Animate, save individual frames, and convert to video using ffmpeg.

        Args:
            fname: Output filename.
            frames: Frames to animate.
            dpi: Dots per inch for saved images.
            framerate: Frame rate of the output video.
            samples: Samples to animate.
            delete_if_exists: Whether to delete existing output file.
            source_path: Source path for temporary files.
            dest_path: Destination path for the output video.
            type: Output video type.
        """
        self._create_temp_dir(path=source_path)
        self.update = True
        self.init()
        frames_list = self._get_indices("frames", frames)
        samples_list = self._get_indices("n_samples", samples)

        try:
            for sample in samples_list:
                self.batch_sample = sample
                for frame in frames_list:
                    self.animate_save(frame, dpi=dpi)
        except Exception as e:
            logging.error("Error during animation: %s", e)
            raise

        self.convert(
            fname,
            delete_if_exists=delete_if_exists,
            framerate=framerate,
            source_path=source_path,
            dest_path=dest_path,
            type=type,
        )

        self._temp_dir.cleanup()

    def convert(
        self,
        fname: str,
        delete_if_exists: bool = False,
        framerate: int = 30,
        source_path: Optional[Union[str, Path]] = None,
        dest_path: Optional[Union[str, Path]] = None,
        type: Literal["mp4", "webm"] = "mp4",
    ) -> None:
        """Convert PNG files in the animations directory to video.

        Args:
            fname: Output filename.
            delete_if_exists: Whether to delete existing output file.
            framerate: Frame rate of the output video.
            source_path: Source path for input PNG files.
            dest_path: Destination path for the output video.
            type: Output video type.
        """
        dest_path = Path(dest_path or self.path)
        dest_path.mkdir(parents=True, exist_ok=True)
        convert(
            source_path or self._path,
            dest_path / f"{fname}.{type}",
            framerate,
            delete_if_exists,
            type=type,
        )


def convert(
    directory: Union[str, Path],
    dest: Union[str, Path],
    framerate: int,
    delete_if_exists: bool,
    type: Literal["mp4", "webm"] = "mp4",
) -> None:
    """Convert PNG files in directory to MP4 or WebM.

    Args:
        directory: Source directory containing PNG files.
        dest: Destination path for the output video.
        framerate: Frame rate of the output video.
        delete_if_exists: Whether to delete existing output file.
        type: Output video type.

    Raises:
        ValueError: If unsupported video type is specified.
        FileExistsError: If output file exists and delete_if_exists is False.
    """
    video = Path(dest)

    if type == "mp4":
        kwargs = dict(
            vcodec="libx264",
            vprofile="high",
            vlevel="4.0",
            vf="pad=ceil(iw/2)*2:ceil(ih/2)*2",  # to make sizes even
            pix_fmt="yuv420p",
            crf=18,
        )
    elif type == "webm":
        kwargs = dict(
            vcodec="libvpx-vp9",
            vf="pad=ceil(iw/2)*2:ceil(ih/2)*2",
            pix_fmt="yuva420p",
            crf=18,
            threads=4,
        )
    else:
        raise ValueError(f"Unsupported video type: {type}")

    if video.exists():
        if delete_if_exists:
            video.unlink()
        else:
            raise FileExistsError(f"File {video} already exists.")

    try:
        (
            ffmpeg.input(f"{directory}/*_*.png", pattern_type="glob", framerate=framerate)
            .output(str(video), **kwargs)
            .run(
                overwrite_output=True,
                quiet=True,
                capture_stdout=True,
                capture_stderr=True,
            )
        )
    except FileNotFoundError as e:
        if "ffmpeg" in str(e):
            logging.warning("Check ffmpeg installation: %s", e)
            return
        else:
            raise
    except ffmpeg.Error as e:
        logging.error("ffmpeg error: %s", e.stderr.decode("utf8"))
        raise e

    logging.info("Created %s", video)


class AnimationCollector(Animation):
    """Collects Animations and updates all axes at once.

    Subclasses must populate the `animations` attribute with Animation objects
    and adhere to the Animation interface.

    Attributes:
        animations (list[Animation]): List of Animation objects to collect.
    """

    animations: list[Animation] = []

    def init(self, frame: int = 0) -> None:
        """Initialize all collected animations.

        Args:
            frame: Initial frame number.
        """
        for animation in self.animations:
            animation.init(frame)
            animation.update = False

    def animate(self, frame: int) -> None:
        """Animate all collected animations for a single frame.

        Args:
            frame: Frame number to animate.
        """
        for animation in self.animations:
            animation.animate(frame)
        if self.update:
            self.update_figure()

    def __setattr__(self, key: str, val: Any) -> None:
        """Set attributes for all Animation objects at once.

        Args:
            key: Attribute name to set.
            val: Value to set for the attribute.
        """
        if key == "batch_sample" and hasattr(self, "animations"):
            for animation in self.animations:
                setattr(animation, key, val)
        super().__setattr__(key, val)



"""Directory classes.

Note:
    Autoreload in jupyter notebooks can break the reference to the directory classes,
    leading to errors after saving of a file that contains a reference to a directory
    class that is then reinstantiated in the jupyter notebook.
    This is why the classes are defined in a separate file. The error can be fixed by
    restarting the kernel.
"""

from datamate import Directory, root

import flyvis

__all__ = ["NetworkDir", "EnsembleDir"]


@root(flyvis.results_dir)
class NetworkDir(Directory):
    """Directory for a network.

    Attributes: Written to by the solver.
        loss (ArrayFile): Loss values over iterations.
        activity (ArrayFile): Mean activity values over iterations.
        activity_min (ArrayFile): Minimum activity values over iterations.
        activity_max (ArrayFile): Maximum activity values over iterations.
        loss_<task> (ArrayFile): Loss values for each specific task over iterations.
        chkpt_index (ArrayFile): Numerical identifiers of checkpoints.
        chkpt_iter (ArrayFile): Iterations at which checkpoints were recorded.
        best_chkpt_index (ArrayFile): Checkpoint index with minimal validation loss.
        dt (ArrayFile): Current time constant of the dataset.
        time_trained (ArrayFile): Total time spent training.

    Attributes: Written by NetworkView.
        __cache__ (Directory): joblib cache.
    """


@root(flyvis.results_dir)
class EnsembleDir(Directory):
    """Contains many NetworkDirs."""



"""Utilities for working with classes."""

from copy import deepcopy
from typing import Any, Dict, Optional, Type
from warnings import warn


def find_subclass(cls: Type, target_subclass_name: str) -> Optional[Type]:
    """
    Recursively search for the target subclass.

    Args:
        cls: The base class to start the search from.
        target_subclass_name: The name of the subclass to find.

    Returns:
        The found subclass, or None if not found.
    """
    for subclass in cls.__subclasses__():
        if subclass.__qualname__ == target_subclass_name:
            return subclass
        # Recursively check the subclasses of the current subclass
        found_subclass = find_subclass(subclass, target_subclass_name)
        if found_subclass is not None:
            return found_subclass
    return None


def forward_subclass(
    cls: Type,
    config: Dict[str, Any] = {},
    subclass_key: str = "type",
    unpack_kwargs: bool = True,
) -> Any:
    """
    Forward to a subclass based on the `<subclass_key>` key in `config`.

    Forwards to the parent class if `<subclass_key>` is not in `config`.

    Args:
        cls: The base class to forward from.
        config: Configuration dictionary containing subclass information.
        subclass_key: Key in the config dictionary specifying the subclass.
        unpack_kwargs: Whether to unpack kwargs when initializing the instance.

    Returns:
        An instance of the specified subclass or the base class.

    Note:
        If the specified subclass is not found, a warning is issued and the base
        class is used instead.
    """
    config = deepcopy(config)
    target_subclass = config.pop(subclass_key, None)

    # Prepare kwargs by removing the subclass_key if it exists
    kwargs = {k: v for k, v in config.items() if k != subclass_key}

    def init_with_kwargs(instance: Any) -> None:
        if unpack_kwargs:
            instance.__init__(**kwargs)
        else:
            instance.__init__(kwargs)

    if target_subclass is not None:
        # Find the target subclass recursively
        subclass = find_subclass(cls, target_subclass)
        if subclass is not None:
            instance = object.__new__(subclass)
            init_with_kwargs(instance)
            return instance
        else:
            warn(
                f"Unrecognized {subclass_key} {target_subclass}. "
                f"Using {cls.__qualname__}.",
                stacklevel=2,
            )
    else:
        warn(f"Missing {subclass_key} in config. Using {cls.__qualname__}.", stacklevel=2)

    # Default case: create an instance of the base class
    instance = object.__new__(cls)
    init_with_kwargs(instance)
    return instance



"""Utility functions for operations on torch.nn.Modules."""

from contextlib import contextmanager
from dataclasses import dataclass
from typing import Generator

from torch import nn


@contextmanager
def simulation(network: nn.Module) -> Generator[None, None, None]:
    """
    Context manager to turn off training mode and require_grad for a network.

    Args:
        network: The neural network module to simulate.

    Yields:
        None

    Example:
        ```python
        model = MyNeuralNetwork()
        with simulation(model):
            # Perform inference or evaluation
            output = model(input_data)
        ```

    Note:
        This context manager temporarily disables gradient computation and sets
        the network to evaluation mode. It restores the original state after
        exiting the context.
    """
    _training = network.training
    network.training = False
    params_require_grad = {}
    for name, p in network.named_parameters():
        params_require_grad[name] = p.requires_grad
        p.requires_grad = False
    try:
        yield
    finally:
        network.training = _training
        for name, p in network.named_parameters():
            p.requires_grad = params_require_grad[name]


@dataclass
class NumberOfParams:
    """
    Dataclass to store the number of free and fixed parameters.

    Attributes:
        free: The number of trainable parameters.
        fixed: The number of non-trainable parameters.
    """

    free: int
    fixed: int


def n_params(nnmodule: nn.Module) -> NumberOfParams:
    """
    Returns the numbers of free and fixed parameters in a PyTorch module.

    Args:
        nnmodule: The PyTorch module to analyze.

    Returns:
        A NumberOfParams object containing the count of free and fixed parameters.

    Example:
        ```python
        model = MyNeuralNetwork()
        param_count = n_params(model)
        print(f"Free parameters: {param_count.free}")
        print(f"Fixed parameters: {param_count.fixed}")
        ```
    """
    n_free = 0
    n_fixed = 0
    for param in nnmodule.parameters():
        if param.requires_grad:
            n_free += param.nelement()
        else:
            n_fixed += param.nelement()
    return NumberOfParams(n_free, n_fixed)



"""Plotting utils."""

from numbers import Number
from typing import Any, Dict, Iterable, List, Literal, Optional, Tuple, Union

import matplotlib as mpl
import matplotlib.colors as colors
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn.functional as F
from matplotlib import colormaps as cm
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colors import ListedColormap, Normalize, TwoSlopeNorm, hex2color
from matplotlib.gridspec import GridSpec

MARKERS = np.array(["o", "^", "s", "*", "+", "h", "p", "8"])


def check_markers(N: int) -> List[str]:
    """
    Check if the number of clusters is larger than the number of markers.

    Args:
        N: Number of clusters.

    Returns:
        List of markers.
    """
    if len(MARKERS) < N:
        return [f"${i}$" for i in range(N)]
    return MARKERS


def get_marker(n: int) -> str:
    """
    Get marker for n.

    Args:
        n: Index of the marker.

    Returns:
        Marker string.
    """
    return check_markers(n)[n]


def init_plot(
    figsize: List[float] = [1, 1],
    title: str = "",
    fontsize: int = 5,
    ax: Axes = None,
    fig: plt.Figure = None,
    projection: str = None,
    set_axis_off: bool = False,
    transparent: bool = False,
    face_alpha: float = 0,
    position: List[float] = None,
    title_pos: Literal["center", "left", "right"] = "center",
    title_y: float = None,
    **kwargs,
) -> Tuple[plt.Figure, Axes]:
    """
    Creates fig and axis object with certain default settings.

    Args:
        figsize: Figure size.
        title: Title of the plot.
        fontsize: Font size for title and labels.
        ax: Existing axis object.
        fig: Existing figure object.
        projection: Projection type (e.g., 'polar').
        set_axis_off: Whether to turn off axis.
        transparent: Whether to make the axis transparent.
        face_alpha: Alpha value for the face color.
        position: Position for newly created axis.
        title_pos: Position of the title.
        title_y: Y-coordinate of the title.

    Returns:
        Tuple containing the figure and axis objects.
    """
    if fig is None:
        fig = plt.figure(figsize=figsize, layout="constrained")
    if ax is not None:
        ax.set_title(title, fontsize=fontsize, loc=title_pos, y=title_y)
    else:
        ax = fig.add_subplot(projection=projection)
        if position is not None:
            ax.set_position(position)
        ax.patch.set_alpha(face_alpha)
        ax.set_title(title, fontsize=fontsize, loc=title_pos, y=title_y)

        if set_axis_off:
            ax.set_axis_off()
    ax.tick_params(axis="both", which="major", labelsize=fontsize)
    if transparent:
        ax.patch.set_alpha(0)
    return fig, ax


def truncate_colormap(
    cmap: colors.Colormap, minval: float = 0.0, maxval: float = 1.0, n: int = 100
) -> colors.LinearSegmentedColormap:
    """
    Truncate colormap.

    Args:
        cmap: Original colormap.
        minval: Minimum value for truncation.
        maxval: Maximum value for truncation.
        n: Number of colors in the new colormap.

    Returns:
        Truncated colormap.
    """
    new_cmap = colors.LinearSegmentedColormap.from_list(
        f"trunc({cmap.name},{minval:.2f},{maxval:.2f})",
        cmap(np.linspace(minval, maxval, n)),
    )
    return new_cmap


def rm_spines(
    ax: Axes,
    spines: Tuple[str, ...] = ("top", "right", "bottom", "left"),
    visible: bool = False,
    rm_xticks: bool = True,
    rm_yticks: bool = True,
) -> Axes:
    """
    Removes spines and ticks from axis.

    Args:
        ax: Matplotlib axis object.
        spines: Tuple of spines to remove.
        visible: Whether to make spines visible.
        rm_xticks: Whether to remove x-ticks.
        rm_yticks: Whether to remove y-ticks.

    Returns:
        Modified axis object.
    """
    for spine in spines:
        ax.spines[spine].set_visible(visible)
    if ("top" in spines or "bottom" in spines) and rm_xticks:
        ax.xaxis.set_ticklabels([])
        ax.xaxis.set_ticks_position("none")
    if ("left" in spines or "right" in spines) and rm_yticks:
        ax.yaxis.set_ticklabels([])
        ax.yaxis.set_ticks_position("none")
    return ax


def get_ax_positions(axes: Iterable[Axes]) -> Tuple[np.ndarray, np.ndarray]:
    """
    Returns the positions of the axes in the figure.

    Args:
        axes: Single ax or iterable of axes.

    Returns:
        Tuple containing arrays of left, bottom, right, and top positions,
        and arrays of centers, widths, and heights.
    """
    axes = np.atleast_1d(axes)
    lefts, bottoms, rights, tops = np.atleast_2d(
        np.array([ax.get_position().extents for ax in axes])
    ).T
    widths = rights - lefts
    heights = tops - bottoms
    centers = np.array([lefts + widths / 2, bottoms + heights / 2])
    return (lefts, bottoms, rights, tops), (centers, widths, heights)


def is_hex(color: str) -> bool:
    """
    Checks if color is hex.

    Args:
        color: Color string.

    Returns:
        True if color is hex, False otherwise.
    """
    return "#" in color


def is_integer_rgb(color: Iterable[int]) -> bool:
    """
    Checks if color is integer RGB.

    Args:
        color: Color tuple or list.

    Returns:
        True if color is integer RGB, False otherwise.
    """
    try:
        return any([c > 1 for c in color])
    except TypeError:
        return False


def get_alpha_colormap(saturated_color: str, number_of_shades: int) -> ListedColormap:
    """
    Create a colormap from a color and a number of shades.

    Args:
        saturated_color: Saturated color string.
        number_of_shades: Number of shades in the colormap.

    Returns:
        ListedColormap object.
    """
    if is_hex(saturated_color):
        rgba = [*hex2color(saturated_color)[:3], 0]
    elif is_integer_rgb(saturated_color):
        rgba = [*list(np.array(saturated_color) / 255.0), 0]

    N = number_of_shades
    colors = []
    alphas = np.linspace(1 / N, 1, N)[::-1]
    for alpha in alphas:
        rgba[-1] = alpha
        colors.append(rgba.copy())

    return ListedColormap(colors)


def polar_to_cmap(
    r: np.ndarray,
    theta: np.ndarray,
    invert: bool = True,
    cmap: colors.Colormap = plt.cm.twilight_shifted,
    norm: Normalize = None,
    sm: ScalarMappable = None,
) -> np.ndarray:
    """
    Maps angle to rgb and amplitude to alpha and returns the resulting array.

    Args:
        r: Amplitude array.
        theta: Angle array.
        invert: Whether to invert the colormap.
        cmap: Colormap.
        norm: Normalization object.
        sm: ScalarMappable object.

    Returns:
        RGBA array.
    """
    sm = sm if sm else plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    r = r / r.max()
    A = np.zeros([theta.shape[0], theta.shape[1], 4])
    RGBA = sm.to_rgba(theta)
    A[:, :, 0] = RGBA[:, :, 0]
    A[:, :, 1] = RGBA[:, :, 1]
    A[:, :, 2] = RGBA[:, :, 2]
    if invert:
        A = 1 - A
        A[:, :, -1] = r  # amplitude
        return A
    else:
        A[:, :, -1] = r  # amplitude
        return A


def add_colorwheel_2d(
    fig: plt.Figure,
    axes: Iterable[Axes] = None,
    pos: Literal[
        "southeast",
        "east",
        "northeast",
        "north",
        "northwest",
        "west",
        "southwest",
        "south",
        "origin",
    ] = "southeast",
    radius: float = 0.25,
    x_offset: float = 0,
    y_offset: float = 0,
    sm: ScalarMappable = None,
    cmap: str = "cm_uniform_2d",
    norm: Normalize = None,
    fontsize: int = 6,
    N: int = 512,
    labelpad: float = 0,
    invert: bool = False,
    mode: Literal["1d", "2d"] = "2d",
    ticks: List[int] = [0, 60, 120],
) -> Tuple[Axes, Axes]:
    """
    Adds a colorwheel to a figure.

    Args:
        fig: Matplotlib figure object.
        axes: Iterable of axes to which the colorwheel will be added.
        pos: Position of the colorwheel.
        radius: Radius of the colorwheel in percentage of the ax radius.
        x_offset: X-offset of the colorwheel in percentage of the cbar diameter.
        y_offset: Y-offset of the colorwheel in percentage of the cbar diameter.
        sm: ScalarMappable object.
        cmap: Colormap name.
        norm: Normalization object.
        fontsize: Font size for tick labels.
        N: Number of samples for the colorwheel.
        labelpad: Padding for tick labels.
        invert: Whether to invert the colormap.
        mode: Mode of the colorwheel ("1d" or "2d").
        ticks: Tick positions in degrees.

    Returns:
        Tuple containing the colorwheel axis and the annotation axis.
    """
    cmap = plt.get_cmap(cmap)

    pos = derive_position_for_supplementary_ax_hex(
        fig,
        axes=axes,
        pos=pos,
        radius=radius,
        x_offset=x_offset,
        y_offset=y_offset,
    )
    cb = fig.add_axes(pos, alpha=0)
    cb.patch.set_alpha(0)

    x = np.linspace(-1, 1, N)
    y = np.linspace(-1, 1, N)
    X, Y = np.meshgrid(x, y)
    R = np.sqrt(X * X + Y * Y)
    circular_mask = R < 1
    R[~circular_mask] = 0
    if mode == "1d":
        R[circular_mask] = 1

    PHI = np.arctan2(Y, X)  # + np.pi
    cb.imshow(
        polar_to_cmap(R, PHI, invert=invert, cmap=cmap, norm=norm, sm=sm),
        origin="lower",
    )
    cb.set_axis_off()

    cs = fig.add_axes(pos, polar=True, label="annotation", alpha=0)
    cs.set_facecolor("none")
    cs.set_yticks([])
    cs.set_yticklabels([])  # turn off radial tick labels (yticks)

    # cosmetic changes to tick labels
    cs.tick_params(pad=labelpad, labelsize=fontsize)
    cs.set_xticks(np.radians(ticks))
    cs.set_xticklabels([x + "°" for x in np.array(ticks).astype(int).astype(str)])

    plt.setp(cs.spines.values(), color="white", linewidth=2)

    return cb, cs


def add_cluster_marker(
    fig: plt.Figure,
    ax: Axes,
    marker: str = "o",
    marker_size: int = 15,
    color: str = "#4F73AE",
    x_offset: float = 0,
    y_offset: float = 0,
) -> None:
    """
    Adds a cluster marker to a figure.

    Args:
        fig: Matplotlib figure object.
        ax: Matplotlib axis object.
        marker: Marker style.
        marker_size: Marker size.
        color: Marker color.
        x_offset: X-offset of the marker in percentage of the ax width.
        y_offset: Y-offset of the marker in percentage of the ax height.
    """
    # make all axes transparent to see the marker regardless where on the figure
    # plane it is
    for _ax in fig.axes:
        _ax.patch.set_alpha(0)

    # create an invisible ax that spans the entire figure to scatter the marker on it
    overlay_ax = [ax for ax in fig.axes if ax.get_label() == "overlay"]
    overlay_ax = (
        overlay_ax[0]
        if overlay_ax
        else fig.add_axes([0, 0, 1, 1], label="overlay", alpha=0)
    )

    overlay_ax.set_ylim(0, 1)
    overlay_ax.set_xlim(0, 1)
    overlay_ax.patch.set_alpha(0)
    rm_spines(overlay_ax, visible=False, rm_xticks=True, rm_yticks=True)

    # get where the axis is actually positioned, that will be annotated with the
    # marker
    left, bottom, width, height = ax.get_position().bounds

    # scatter the marker relative to that position of the ax
    overlay_ax.scatter(
        left + x_offset * width,
        bottom + y_offset * height,
        marker=marker,
        s=marker_size,
        color=color,
    )


def derive_position_for_supplementary_ax(
    fig: plt.Figure,
    pos: Literal["right", "left", "top", "bottom"] = "right",
    width: float = 0.04,
    height: float = 0.5,
    x_offset: float = 0,
    y_offset: float = 0,
    axes: Iterable[Axes] = None,
) -> List[float]:
    """
    Returns a position for a supplementary ax.

    Args:
        fig: Matplotlib figure object.
        pos: Position of the supplementary ax relative to the main axes.
        width: Width of the supplementary ax in percentage of the main ax width.
        height: Height of the supplementary ax in percentage of the main ax height.
        x_offset: X-offset of the supplementary ax in percentage of the main ax width.
        y_offset: Y-offset of the supplementary ax in percentage of the main ax height.
        axes: Iterable of axes to which the supplementary ax will be added.

    Returns:
        List containing the left, bottom, width, and height of the supplementary ax.
    """
    axes = axes if axes is not None else fig.get_axes()
    x0, y0, x1, y1 = np.array([ax.get_position().extents for ax in axes]).T
    x0, y0, x1, y1 = x0.min(), y0.min(), x1.max(), y1.max()
    ax_width = x1 - x0
    ax_height = y1 - y0
    positions = {
        "right": [
            x1 + ax_width * width / 2 + ax_width * width * x_offset,  # left
            y0 + (1 - height) * ax_height / 2 + ax_height * height * y_offset,  # bottom
            ax_width * width,  # width
            ax_height * height,  # height
        ],
        "left": [
            x0 - 3 / 2 * ax_width * width + ax_width * width * x_offset,  # left
            y0 + (1 - height) * ax_height / 2 + y_offset,  # bottom
            ax_width * width,  # width
            ax_height * height,  # height
        ],
        "top": [
            x1
            - ax_width * width
            + ax_width * width * x_offset,  # x0 + (1 - width) * ax_width/2
            y1 + ax_height * height / 2 + ax_height * height * y_offset,
            ax_width * width,
            ax_height * height,
        ],
        "bottom": [
            x0 + (1 - width) * ax_width / +ax_width * width * x_offset,
            y0 - 3 / 2 * ax_height * height + ax_height * height * y_offset,
            ax_width * width,
            ax_height * height,
        ],
    }
    return positions[pos]


def derive_position_for_supplementary_ax_hex(
    fig: plt.Figure,
    axes: Iterable[Axes] = None,
    pos: Literal[
        "southeast",
        "east",
        "northeast",
        "north",
        "northwest",
        "west",
        "southwest",
        "south",
        "origin",
    ] = "southwest",
    radius: float = 0.25,
    x_offset: float = 0,
    y_offset: float = 0,
) -> List[float]:
    """
    Returns a position for a supplementary ax.

    Args:
        fig: Matplotlib figure object.
        axes: Iterable of axes to which the supplementary ax will be added.
        pos: Position of the supplementary ax relative to the main axes.
        radius: Radius of the supplementary ax in percentage of the main ax radius.
        x_offset: X-offset of the supplementary ax in percentage of the main ax width.
        y_offset: Y-offset of the supplementary ax in percentage of the main ax height.

    Returns:
        List containing the left, bottom, width, and height of the supplementary ax.
    """
    axes = axes if axes is not None else fig.get_axes()
    x0, y0, x1, y1 = np.array([ax.get_position().extents for ax in axes]).T
    x0, y0, x1, y1 = x0.min(), y0.min(), x1.max(), y1.max()
    axes_width = x1 - x0
    axes_height = y1 - y0
    axes_radius = (axes_width + axes_height) / 4
    new_ax_radius = axes_radius * radius
    position = {
        "southeast": [
            x1 + x_offset * 2 * new_ax_radius,
            y1
            - axes_height
            - 2 * new_ax_radius
            + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "east": [
            x1 + x_offset * 2 * new_ax_radius,
            y1
            - axes_height / 2
            - new_ax_radius
            + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "northeast": [
            x1 + x_offset * 2 * new_ax_radius,
            y1 + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "north": [
            x1 - axes_width / 2 - new_ax_radius + x_offset * 2 * new_ax_radius,
            y1 + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "northwest": [
            x1 - axes_width - 2 * new_ax_radius + x_offset * 2 * new_ax_radius,
            y1 + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "west": [
            x1 - axes_width - 2 * new_ax_radius + x_offset * 2 * new_ax_radius,
            y1
            - axes_height / 2
            - new_ax_radius
            + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "southwest": [
            x1 - axes_width - 2 * new_ax_radius + x_offset * 2 * new_ax_radius,
            y1
            - axes_height
            - 2 * new_ax_radius
            + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "south": [
            x1 - axes_width / 2 - new_ax_radius + x_offset * 2 * new_ax_radius,
            y1
            - axes_height
            - 2 * new_ax_radius
            + y_offset * 2 * new_ax_radius,  # + radius ** 2 * axes_height,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
        "origin": [
            x0 + x_offset,
            y0 + y_offset,
            2 * new_ax_radius,
            2 * new_ax_radius,
        ],
    }
    return position[pos]


def add_colorbar_to_fig(
    fig: plt.Figure,
    axes: Iterable[Axes] = None,
    pos: Literal["right", "left", "top", "bottom"] = "right",
    width: float = 0.04,
    height: float = 0.5,
    x_offset: float = 0,
    y_offset: float = 0,
    cmap: colors.Colormap = cm.get_cmap("binary"),
    fontsize: int = 10,
    tick_length: float = 1.5,
    tick_width: float = 0.75,
    rm_outline: bool = True,
    ticks: Iterable[float] = None,
    norm: Normalize = None,
    label: str = "",
    plain: bool = False,
    use_math_text: bool = False,
    scilimits: Tuple[float, float] = None,
    style: str = "",
    alpha: float = 1,
    n_ticks: int = 9,  # only effective if norm is TwoSlopeNorm
    discrete: bool = False,
    n_discrete: int = None,
    discrete_labels: Iterable[str] = None,
    n_decimals: int = 2,
) -> mpl.colorbar.Colorbar:
    """
    Adds a colorbar to a figure.

    Args:
        fig: Matplotlib figure object.
        axes: Iterable of axes to which the colorbar will be added.
        pos: Position of the colorbar.
        width: Width of the colorbar in percentage of the ax width.
        height: Height of the colorbar in percentage of the ax height.
        x_offset: X-offset of the colorbar in percentage of the ax width.
        y_offset: Y-offset of the colorbar in percentage of the ax height.
        cmap: Colormap.
        fontsize: Font size for tick labels.
        tick_length: Length of the tick marks.
        tick_width: Width of the tick marks.
        rm_outline: Whether to remove the outline of the colorbar.
        ticks: Tick positions.
        norm: Normalization object.
        label: Colorbar label.
        plain: Whether to remove tick labels.
        use_math_text: Whether to use math text for tick labels.
        scilimits: Limits for scientific notation.
        style: Style for scientific notation.
        alpha: Alpha value for the colorbar.
        n_ticks: Number of ticks for TwoSlopeNorm.
        discrete: Whether to use discrete colors.
        n_discrete: Number of discrete colors.
        discrete_labels: Labels for discrete colors.
        n_decimals: Number of decimal places for tick labels.

    Returns:
        Matplotlib colorbar object.
    """
    _orientation = "vertical" if pos in ("left", "right") else "horizontal"

    position = derive_position_for_supplementary_ax(
        fig=fig,
        pos=pos,
        width=width,
        height=height,
        x_offset=x_offset,
        y_offset=y_offset,
        axes=axes,
    )
    cbax = fig.add_axes(position, label="cbar")

    cbar = mpl.colorbar.ColorbarBase(
        cbax,
        cmap=cmap,
        norm=norm,
        orientation=_orientation,
        ticks=ticks,
        alpha=alpha,
    )
    cbar.set_label(fontsize=fontsize, label=label)
    cbar.ax.tick_params(labelsize=fontsize, length=tick_length, width=tick_width)
    if pos in ("left", "right"):
        scalarformatter = isinstance(
            cbar.ax.yaxis.get_major_formatter(), mpl.ticker.ScalarFormatter
        )
        cbar.ax.yaxis.set_ticks_position(pos)
        cbar.ax.yaxis.set_label_position(pos)
        cbar.ax.yaxis.get_offset_text().set_fontsize(fontsize)
        cbar.ax.yaxis.get_offset_text().set_horizontalalignment("left")
        cbar.ax.yaxis.get_offset_text().set_verticalalignment("bottom")
    else:
        scalarformatter = isinstance(
            cbar.ax.xaxis.get_major_formatter(), mpl.ticker.ScalarFormatter
        )
        cbar.ax.xaxis.set_ticks_position(pos)
        cbar.ax.xaxis.set_label_position(pos)
        cbar.ax.xaxis.get_offset_text().set_fontsize(fontsize)
        cbar.ax.xaxis.get_offset_text().set_verticalalignment("top")
        cbar.ax.xaxis.get_offset_text().set_horizontalalignment("left")

    if scalarformatter:
        cbar.ax.ticklabel_format(
            style=style, useMathText=use_math_text, scilimits=scilimits
        )

    if rm_outline:
        cbar.outline.set_visible(False)

    if isinstance(norm, TwoSlopeNorm):
        vmin = norm.vmin
        vmax = norm.vmax
        vcenter = norm.vcenter
        left_ticks = np.linspace(vmin, vcenter, n_ticks // 2)
        right_ticks = np.linspace(vcenter, vmax, n_ticks // 2)
        ticks = ticks or [*left_ticks, *right_ticks[1:]]
        cbar.set_ticks(
            ticks,
            labels=[f"{t:.{n_decimals}f}" for t in ticks],
            fontsize=fontsize,
        )

    if plain:
        cbar.set_ticks([])

    if discrete:
        # to put ticklabels for discrete colors in the middle
        if not n_discrete:
            raise ValueError(f"n_discrete {n_discrete}")
        lim = cbar.ax.get_ylim() if pos in ["left", "right"] else cbar.ax.get_xlim()
        color_width = (lim[1] - lim[0]) / n_discrete
        label_offset_to_center = color_width / 2
        labels = np.arange(n_discrete)
        loc = np.linspace(*lim, n_discrete, endpoint=False) + label_offset_to_center
        cbar.set_ticks(loc)
        cbar.set_ticklabels(discrete_labels or labels)

    return cbar


def get_norm(
    norm: Normalize = None,
    vmin: float = None,
    vmax: float = None,
    midpoint: float = None,
    log: bool = None,
    symlog: float = None,
) -> Normalize:
    """
    Returns a normalization object for color normalization.

    Args:
        norm: A class which, when called, can normalize data into an interval
            [vmin, vmax].
        vmin: Minimum value for normalization.
        vmax: Maximum value for normalization.
        midpoint: Midpoint value so that data is normalized around it.
        log: Whether to normalize on a log-scale.
        symlog: Normalizes to symlog with linear range around the range (-symlog, symlog).

    Returns:
        Normalization object.
    """
    if norm:
        return norm

    if all(val is not None for val in (vmin, vmax)):
        vmin -= 1e-15
        vmax += 1e-15

    if all(val is not None for val in (vmin, vmax, midpoint)):
        if vmin > midpoint or np.isclose(vmin, midpoint, atol=1e-9):
            vmin = midpoint - vmax
        if vmax < midpoint or np.isclose(vmax, midpoint, atol=1e-9):
            vmax = midpoint - vmin
        return TwoSlopeNorm(vcenter=midpoint, vmin=vmin, vmax=vmax)
    elif all(val is not None for val in (vmin, vmax, log)):
        return mpl.colors.LogNorm(vmin=vmin, vmax=vmax)
    elif all(val is not None for val in (vmin, vmax, symlog)):
        v = max(np.abs(vmin), np.abs(vmax))
        return mpl.colors.SymLogNorm(symlog, vmin=-v, vmax=v)
    elif all(val is not None for val in (vmin, vmax)):
        return mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    else:
        return None


def get_scalarmapper(
    scalarmapper: ScalarMappable = None,
    cmap: colors.Colormap = None,
    norm: Normalize = None,
    vmin: float = None,
    vmax: float = None,
    midpoint: float = None,
    log: bool = None,
    symlog: float = None,
) -> Tuple[ScalarMappable, Normalize]:
    """
    Returns scalarmappable with norm from `get_norm` and cmap.

    Args:
        scalarmapper: Scalarmappable for data to RGBA mapping.
        cmap: Colormap.
        norm: Normalization object.
        vmin: Minimum value for normalization.
        vmax: Maximum value for normalization.
        midpoint: Midpoint value for normalization.
        log: Whether to normalize on a log-scale.
        symlog: Normalizes to symlog with linear range around the range (-symlog, symlog).

    Returns:
        Tuple containing the scalarmappable and the normalization object.
    """
    if scalarmapper:
        return scalarmapper, norm

    norm = get_norm(
        norm=norm,
        vmin=vmin,
        vmax=vmax,
        midpoint=midpoint,
        log=log,
        symlog=symlog,
    )
    return plt.cm.ScalarMappable(norm=norm, cmap=cmap), norm


def get_lims(
    z: Union[np.ndarray, Iterable[np.ndarray]],
    offset: float,
    min: float = None,
    max: float = None,
) -> Tuple[float, float]:
    """
    Get scalar bounds of Ndim-array-like structure with relative offset.

    Args:
        z: Ndim-array-like structure.
        offset: Relative offset for the bounds.
        min: Minimum value for the bounds.
        max: Maximum value for the bounds.

    Returns:
        Tuple containing the minimum and maximum values.
    """

    def sub_nan(val: float, sub: float) -> float:
        if np.isnan(val):
            return sub
        else:
            return val

    if isinstance(z, (tuple, list)):
        z = list(map(lambda x: get_lims(x, offset), z))
    if isinstance(z, torch.Tensor):
        z = z.detach().cpu().numpy()
    z = np.array(z)[~np.isinf(z)]
    if not z.any():
        return -1, 1
    _min, _max = np.nanmin(z), np.nanmax(z)
    _range = np.abs(_max - _min)
    _min -= _range * offset
    _max += _range * offset
    _min, _max = sub_nan(_min, 0), sub_nan(_max, 1)
    if min is not None:
        _min = np.min((min, _min))
    if max is not None:
        _max = np.max((max, _max))
    return _min, _max


def avg_pool(trace: np.ndarray, N: int) -> np.ndarray:
    """
    Smoothes (multiple) traces over the second dimension using the GPU.

    Args:
        trace: Array of shape (N, t).
        N: Window size for averaging.

    Returns:
        Smoothed trace array.
    """
    shape = trace.shape
    trace = trace.reshape(np.prod(shape[:-1]), 1, shape[-1])
    with torch.no_grad():
        trace_smooth = (
            F.avg_pool1d(torch.tensor(trace, dtype=torch.float32), N, N).cpu().numpy()
        )
    return trace_smooth.reshape(shape[0], -1)


def width_n_height(
    N: int, aspect_ratio: float, max_width: int = None, max_height: int = None
) -> Tuple[int, int]:
    """
    Integer width and height for a grid of N plots with aspect ratio.

    Args:
        N: Number of plots.
        aspect_ratio: Aspect ratio of the grid.
        max_width: Maximum width of the grid.
        max_height: Maximum height of the grid.

    Returns:
        Tuple containing the width and height of the grid.
    """
    if max_width is not None and max_height is not None:
        raise ValueError

    _sqrt = int(np.ceil(np.sqrt(N)))

    gridwidth = np.ceil(_sqrt * aspect_ratio).astype(int)
    gridheight = np.ceil(_sqrt / aspect_ratio).astype(int)

    gridwidth = max(1, min(N, gridwidth, np.ceil(N / gridheight)))
    gridheight = max(1, min(N, gridheight, np.ceil(N / gridwidth)))

    if max_width is not None and gridwidth > max_width:
        gridwidth = max_width
        gridheight = np.ceil(N / gridwidth)

    if max_height is not None and gridheight > max_height:
        gridheight = max_height
        gridwidth = np.ceil(N / gridheight)

    assert gridwidth * gridheight >= N

    return int(gridwidth), int(gridheight)


def get_axis_grid(
    alist: Iterable = None,
    gridwidth: int = None,
    gridheight: int = None,
    max_width: int = None,
    max_height: int = None,
    fig: plt.Figure = None,
    ax: Axes = None,
    axes: Iterable[Axes] = None,
    aspect_ratio: float = 1,
    figsize: List[float] = None,
    scale: Union[int, Iterable[int]] = 3,
    projection: Union[str, Iterable[str]] = None,
    as_matrix: bool = False,
    fontsize: int = 5,
    wspace: float = 0.1,
    hspace: float = 0.3,
    alpha: float = 1,
    sharex: Axes = None,
    sharey: Axes = None,
    unmask_n: int = None,
) -> Tuple[plt.Figure, Union[List[Axes], np.ndarray], Tuple[int, int]]:
    """
    Create axis grid for a list of elements or integer width and height.

    Args:
        alist: List of elements to create grid for.
        gridwidth: Width of grid.
        gridheight: Height of grid.
        max_width: Maximum width of grid.
        max_height: Maximum height of grid.
        fig: Existing figure to use.
        ax: Existing axis to use. This ax will be divided into a grid of axes with the
            same size as the grid.
        axes: Existing axes to use.
        aspect_ratio: Aspect ratio of grid.
        figsize: Figure size.
        scale: Scales figure size by this factor(s) times the grid width and height.
        projection: Projection of axes.
        as_matrix: Return axes as matrix.
        fontsize: Fontsize of axes.
        wspace: Width space between axes.
        hspace: Height space between axes.
        alpha: Alpha of axes.
        sharex: Share x axis. Only effective if a new grid of axes is created.
        sharey: Share y axis. Only effective if a new grid of axes is created.
        unmask_n: Number of elements to unmask. If None, all elements are unmasked.
            If provided elements at indices >= unmask_n are padded with nans.

    Returns:
        Tuple containing the figure, axes, and the grid width and height.
    """
    if alist is not None and (
        gridwidth is None or gridheight is None or gridwidth * gridheight != len(alist)
    ):
        gridwidth, gridheight = width_n_height(
            len(alist), aspect_ratio, max_width=max_width, max_height=max_height
        )
    elif gridwidth and gridheight:
        alist = range(gridwidth * gridheight)
    else:
        raise ValueError("Either specify alist or gridwidth and gridheight manually.")
    unmask_n = unmask_n or len(alist)
    if figsize is not None:
        pass
    elif isinstance(scale, Number):
        figsize = [scale * gridwidth, scale * gridheight]
    elif isinstance(scale, Iterable) and len(scale) == 2:
        figsize = [scale[0] * gridwidth, scale[1] * gridheight]

    if fig is None:
        fig = figure(figsize=figsize, hspace=hspace, wspace=wspace)

    if not isinstance(projection, Iterable) or isinstance(projection, str):
        projection = (projection,) * len(alist)

    if isinstance(ax, Iterable):
        assert len(ax) == len(alist)
        if isinstance(ax, dict):
            ax = list(ax.values())
        return fig, ax, (gridwidth, gridheight)

    if ax:
        # divide an existing ax in a figure
        matrix = np.ones(gridwidth * gridheight) * np.nan
        for i in range(len(alist)):
            matrix[i] = i
        axes = divide_axis_to_grid(
            ax,
            matrix=matrix.reshape(gridwidth, gridheight),
            wspace=wspace,
            hspace=hspace,
        )
        axes = list(axes.values())
    elif axes is None:
        # fill a figure with axes
        axes = []
        _sharex, _sharey = None, None

        for i, _ in enumerate(alist):
            if i < unmask_n:
                ax = subplot(
                    "",
                    grid=(gridheight, gridwidth),
                    location=(int(i // gridwidth), int(i % gridwidth)),
                    rowspan=1,
                    colspan=1,
                    sharex=_sharex,
                    sharey=_sharey,
                    projection=projection[i],
                )

                if sharex is not None:
                    sharex = ax
                if sharey is not None:
                    sharey = ax

                axes.append(ax)
            else:
                axes.append(np.nan)

    for ax in axes:
        if isinstance(ax, Axes):
            ax.tick_params(axis="both", which="major", labelsize=fontsize)
            ax.patch.set_alpha(alpha)

    if as_matrix:
        axes = np.array(axes).reshape(gridheight, gridwidth)

    return fig, axes, (gridwidth, gridheight)


def figure(
    figsize: List[float],
    hspace: float = 0.3,
    wspace: float = 0.1,
    left: float = 0.125,
    right: float = 0.9,
    top: float = 0.9,
    bottom: float = 0.1,
    frameon: bool = None,
) -> plt.Figure:
    """
    Create a figure with the given size and spacing.

    Args:
        figsize: Figure size.
        hspace: Height space between subplots.
        wspace: Width space between subplots.
        left: Left margin.
        right: Right margin.
        top: Top margin.
        bottom: Bottom margin.
        frameon: Whether to draw the figure frame.

    Returns:
        Matplotlib figure object.
    """
    fig = plt.figure(figsize=figsize, frameon=frameon)
    plt.subplots_adjust(
        hspace=hspace,
        wspace=wspace,
        left=left,
        top=top,
        right=right,
        bottom=bottom,
    )
    return fig


def subplot(
    title: str = "",
    grid: Tuple[int, int] = (1, 1),
    location: Tuple[int, int] = (0, 0),
    colspan: int = 1,
    rowspan: int = 1,
    projection: str = None,
    sharex: Axes = None,
    sharey: Axes = None,
    xlabel: str = "",
    ylabel: str = "",
    face_alpha: float = 1.0,
    fontisze: int = 5,
    title_pos: Literal["center", "left", "right"] = "center",
    position: List[float] = None,
    **kwargs,
) -> Axes:
    """
    Create a subplot using subplot2grid with some extra options.

    Args:
        title: Title of the subplot.
        grid: Grid shape.
        location: Location of the subplot in the grid.
        colspan: Number of columns the subplot spans.
        rowspan: Number of rows the subplot spans.
        projection: Projection type (e.g., 'polar').
        sharex: Axis to share x-axis with.
        sharey: Axis to share y-axis with.
        xlabel: X-axis label.
        ylabel: Y-axis label.
        face_alpha: Alpha value for the face color.
        fontisze: Font size for title and labels.
        title_pos: Position of the title.
        position: Position for the subplot.

    Returns:
        Matplotlib axis object.
    """
    ax = plt.subplot2grid(
        grid,
        location,
        colspan,
        rowspan,
        sharex=sharex,
        sharey=sharey,
        projection=projection,
    )
    if position:
        ax.set_position(position)
    ax.patch.set_alpha(face_alpha)

    ax.set_title(title, fontsize=fontisze, loc=title_pos)

    plt.xlabel(xlabel, fontsize=fontisze)
    plt.ylabel(ylabel, fontsize=fontisze)

    return ax


def divide_axis_to_grid(
    ax: Axes,
    matrix: np.ndarray = ((0, 1, 2), (3, 3, 3)),
    wspace: float = 0.1,
    hspace: float = 0.1,
    projection: str = None,
) -> Dict[Any, Axes]:
    """
    Divides an existing axis inside a figure to a grid specified by unique elements
    in a matrix.

    Args:
        ax: Existing Axes object.
        matrix: Grid matrix, where each unique element specifies a new axis.
        wspace: Horizontal space between new axes.
        hspace: Vertical space between new axes.
        projection: Projection of new axes.

    Returns:
        Dictionary of new axes, where keys are unique elements in the matrix.

    Example:
        ```python
        fig = plt.figure()
        ax = plt.subplot()
        plt.tight_layout()
        divide_axis_to_grid(ax, matrix=[[0, 1, 1, 1, 2, 2, 2],
                                            [3, 4, 5, 6, 2, 2, 2],
                                            [3, 7, 7, 7, 2, 2, 2],
                                            [3, 8, 8, 12, 2, 2, 2],
                                            [3, 10, 11, 12, 2, 2, 2]],
                                wspace=0.1, hspace=0.1)
        ```
    """

    # get position of original axis, and dispose it
    x0, y0, x1, y1 = ax.get_position().extents
    ax.set_axis_off()
    ax.patch.set_alpha(0)
    fig = ax.figure

    # get grid shape
    n_row, n_col = np.array(matrix).shape

    # get geometry params
    width = x1 - x0
    height = y1 - y0
    height_per_row = height / n_row
    width_per_col = width / n_col

    _ax_pos = {}

    for i, row in enumerate(matrix):
        for j, _ax in enumerate(row):
            # get occurence of unique element per row and column
            _ax_per_row = sum([1 for _ in np.array(matrix).T[j] if _ == _ax])
            _ax_per_col = sum([1 for _ in row if _ == _ax])

            # compute positioning of _ax
            left = x0 + j * width_per_col + wspace / 2
            bottom = y0 + height - (i + _ax_per_row) * height_per_row + hspace / 2
            _width = width_per_col * _ax_per_col - min(
                wspace / 2, width_per_col * _ax_per_col
            )
            _height = height_per_row * _ax_per_row - min(
                hspace / 2, height_per_row * _ax_per_row
            )

            # store positioning
            if _ax not in _ax_pos and not np.isnan(_ax):
                _ax_pos[_ax] = [left, bottom, _width, _height]

    # add axis to existing figure and store in dict
    axes = {k: None for k in _ax_pos}
    for _ax, pos in _ax_pos.items():
        axes[_ax] = fig.add_axes(pos, projection=projection)

    return axes


def divide_figure_to_grid(
    matrix: List[List[int]] = [
        [0, 1, 1, 1, 2, 2, 2],
        [3, 4, 5, 6, 2, 2, 2],
        [3, 7, 7, 7, 2, 2, 2],
        [3, 8, 8, 12, 2, 2, 2],
        [3, 10, 11, 12, 2, 2, 2],
    ],
    as_matrix: bool = False,
    alpha: float = 0,
    constrained_layout: bool = False,
    fig: Optional[plt.Figure] = None,
    figsize: List[float] = [10, 10],
    projection: Optional[Union[str, List[str]]] = None,
    wspace: float = 0.1,
    hspace: float = 0.3,
    no_spines: bool = False,
    keep_nan_axes: bool = False,
    fontsize: int = 5,
    reshape_order: Literal["C", "F", "A", "K"] = "F",
) -> Tuple[plt.Figure, Union[Dict[int, plt.Axes], np.ndarray]]:
    """
    Creates a figure grid specified by the arrangement of unique elements in a matrix.

    Info:
        `matplotlib` now also has `matplotlib.pyplot.subplot_mosaic` which does the
        same thing and should be used instead.

    Args:
        matrix: Grid layout specification.
        as_matrix: If True, return axes as a numpy array.
        alpha: Alpha value for axis patches.
        constrained_layout: Use constrained layout for the figure.
        fig: Existing figure to use. If None, a new figure is created.
        figsize: Figure size in inches.
        projection: Projection type for the axes.
        wspace: Width space between subplots.
        hspace: Height space between subplots.
        no_spines: If True, remove spines from all axes.
        keep_nan_axes: If True, keep axes for NaN values in the matrix.
        fontsize: Font size for tick labels.
        reshape_order: Order to use when reshaping the axes array.

    Returns:
        A tuple containing the figure and a dictionary or numpy array of axes.
    """

    def _array_to_slice(array):
        step = 1
        start = array.min()
        stop = array.max() + 1
        return slice(start, stop, step)

    fig = plt.figure(figsize=figsize) if fig is None else fig
    fig.set_constrained_layout(constrained_layout)
    matrix = np.ma.masked_invalid(matrix)
    rows, columns = matrix.shape

    gs = GridSpec(rows, columns, figure=fig, hspace=hspace, wspace=wspace)

    axes = {}
    for val in np.unique(matrix[~matrix.mask]):
        _row_ind, _col_ind = np.where(matrix == val)
        _row_slc, _col_slc = _array_to_slice(_row_ind), _array_to_slice(_col_ind)

        _projection = projection[val] if isinstance(projection, dict) else projection
        ax = fig.add_subplot(gs[_row_slc, _col_slc], projection=_projection)
        ax.patch.set_alpha(alpha)
        ax.tick_params(axis="both", which="major", labelsize=fontsize)
        if projection is None:
            ax.spines["top"].set_visible(False)
            ax.spines["right"].set_visible(False)
        axes[val] = ax

    if keep_nan_axes:
        for _row_ind, _col_ind in np.array(np.where(np.isnan(matrix))).T:
            _row_slc, _col_slc = _array_to_slice(_row_ind), _array_to_slice(_col_ind)
            ax = fig.add_subplot(gs[_row_slc, _col_slc])
            ax.patch.set_alpha(alpha)
            rm_spines(ax, ("left", "right", "top", "bottom"))

    if no_spines:
        for ax in axes.values():
            rm_spines(ax, rm_xticks=True, rm_yticks=True)

    if as_matrix:
        if reshape_order == "special":
            # reshape based on elements in matrix
            ax_matrix = np.ones(matrix.shape, dtype=object) * np.nan
            for key, value in axes.items():
                _row_ind, _col_ind = np.where(matrix == key)
                ax_matrix[_row_ind, _col_ind] = value
            axes = ax_matrix
        else:
            # reshape without considering specific locations
            _axes = np.ones(matrix.shape, dtype=object).flatten() * np.nan
            _axes[: len(axes)] = np.array(list(axes.values()), dtype=object)
            axes = _axes.reshape(matrix.shape, order=reshape_order)

    return fig, axes


def scale(
    x: np.ndarray,
    y: np.ndarray,
    wpad: float = 0.1,
    hpad: float = 0.1,
    wspace: float = 0,
    hspace: float = 0,
) -> Tuple[np.ndarray, np.ndarray, float, float]:
    """
    Scale x and y coordinates to fit within a specified padding and spacing.

    Args:
        x: Array of x-coordinates.
        y: Array of y-coordinates.
        wpad: Width padding.
        hpad: Height padding.
        wspace: Width space between elements.
        hspace: Height space between elements.

    Returns:
        A tuple containing scaled x, y coordinates, width, and height.
    """
    x, y = np.array(x), np.array(y)
    assert len(x) == len(y)

    # Min-Max Scale x-Positions.
    width = (1 - 2 * wpad) / (2 * np.ceil(np.median(np.unique(x))))
    width = width - wspace * width
    x = (x - np.min(x)) / (np.max(x) + np.min(x)) * (1 - width) * (1 - wpad * 2) + wpad

    # Min-Max Scale y-Positions.
    height = (1 - 2 * hpad) / (2 * np.ceil(np.median(np.unique(y))))
    height = height - hspace * height
    y = (y - np.min(y)) / (np.max(y) + np.min(y)) * (1 - height) * (1 - hpad * 2) + hpad
    return x, y, width, height


def ax_scatter(
    x: np.ndarray,
    y: np.ndarray,
    fig: Optional[plt.Figure] = None,
    figsize: List[float] = [7, 7],
    hspace: float = 0,
    wspace: float = 0,
    hpad: float = 0.1,
    wpad: float = 0.1,
    alpha: float = 0,
    zorder: int = 10,
    projection: Optional[str] = None,
    labels: Optional[List[str]] = None,
) -> Tuple[plt.Figure, List[plt.Axes], List[List[float]]]:
    """
    Creates scattered axes in a given or new figure.

    Args:
        x: Array of x-coordinates.
        y: Array of y-coordinates.
        fig: Existing figure to use. If None, a new figure is created.
        figsize: Figure size in inches.
        hspace: Height space between subplots.
        wspace: Width space between subplots.
        hpad: Height padding.
        wpad: Width padding.
        alpha: Alpha value for axis patches.
        zorder: Z-order for axis patches.
        projection: Projection type for the axes.
        labels: List of labels for each axis.

    Returns:
        A tuple containing the figure, a list of axes, and a list of center coordinates.
    """
    x, y, width, height = scale(x, y, wpad, hpad, wspace, hspace)

    # Create axes in figure.
    fig = fig or plt.figure(figsize=figsize)
    axes = []
    for i, (_x, _y) in enumerate(zip(x, y)):
        ax = fig.add_axes(
            [_x, _y, width, height],
            projection=projection,
            label=labels[i] if labels is not None else None,
        )
        ax.set_zorder(zorder)
        ax.patch.set_alpha(alpha)
        axes.append(ax)

    center = []
    for ax in axes:
        _, (_center, _, _) = get_ax_positions(ax)
        center.append(_center.flatten().tolist())
    return fig, axes, center


def color_labels(labels: List[str], color, ax):
    for label in labels:
        color_label(label, color, ax)


def color_label(
    label: str, color: Union[str, Tuple[float, float, float]], ax: plt.Axes
) -> None:
    """
    Color a specific label in the given axes.

    Args:
        label: The label text to color.
        color: The color to apply to the label.
        ax: The matplotlib axes object.
    """
    for t in ax.texts:
        if t.get_text() == label:
            t.set_color(color)

    for tick in ax.xaxis.get_major_ticks():
        if tick.label1.get_text() == label:
            tick.label1.set_color(color)

    for tick in ax.yaxis.get_major_ticks():
        if tick.label1.get_text() == label:
            tick.label1.set_color(color)

    if ax.xaxis.get_label().get_text() == label:
        ax.xaxis.get_label().set_color(color)

    if ax.yaxis.get_label().get_text() == label:
        ax.yaxis.get_label().set_color(color)


def boldify_labels(labels: List[str], ax: plt.Axes) -> None:
    """
    Make specific labels bold in the given axes.

    Args:
        labels: List of label texts to make bold.
        ax: The matplotlib axes object.
    """
    for t in ax.texts:
        if t.get_text() in labels:
            t.set_weight("bold")

    for tick in ax.xaxis.get_major_ticks():
        if tick.label1.get_text() in labels:
            tick.label1.set_weight("bold")

    for tick in ax.yaxis.get_major_ticks():
        if tick.label1.get_text() in labels:
            tick.label1.set_weight("bold")

    if ax.xaxis.get_label().get_text() in labels:
        ax.xaxis.get_label().set_weight("bold")

    if ax.yaxis.get_label().get_text() in labels:
        ax.yaxis.get_label().set_weight("bold")


def scatter_on_violins_or_bars(
    data: np.ndarray,
    ax: Axes,
    xticks: Optional[np.ndarray] = None,
    indices: Optional[np.ndarray] = None,
    s: float = 5,
    zorder: int = 100,
    facecolor: Union[
        str, Tuple[float, float, float], List[Union[str, Tuple[float, float, float]]]
    ] = "none",
    edgecolor: Union[
        str, Tuple[float, float, float], List[Union[str, Tuple[float, float, float]]]
    ] = "k",
    linewidth: float = 0.5,
    alpha: float = 0.35,
    uniform: List[float] = [-0.35, 0.35],
    seed: int = 42,
    marker: str = "o",
    **kwargs,
) -> None:
    """
    Scatter data points on violin or bar plots.

    Args:
        data: Array of shape (n_samples, n_random_variables).
        ax: Matplotlib axes object to plot on.
        xticks: X-axis tick positions.
        indices: Selection along sample dimension.
        s: Marker size.
        zorder: Z-order for plotting.
        facecolor: Color(s) for marker face.
        edgecolor: Color(s) for marker edge.
        linewidth: Width of marker edge.
        alpha: Transparency of markers.
        uniform: Range for uniform distribution of x-positions.
        seed: Random seed for reproducibility.
        marker: Marker style.
        **kwargs: Additional keyword arguments for plt.scatter.

    Returns:
        None
    """
    random = np.random.RandomState(seed)

    if xticks is None:
        xticks = ax.get_xticks()
    data = np.atleast_2d(data)
    indices = indices if indices is not None else range(data.shape[0])

    if (
        not isinstance(facecolor, Iterable)
        or len(facecolor) != len(data)
        or isinstance(facecolor, str)
    ):
        facecolor = (facecolor,) * len(indices)

    if (
        not isinstance(edgecolor, Iterable)
        or len(edgecolor) != len(data)
        or isinstance(edgecolor, str)
    ):
        edgecolor = (edgecolor,) * len(indices)

    for i, model_index in enumerate(indices):
        ax.scatter(
            xticks + random.uniform(*uniform, size=len(xticks)),
            data[model_index],
            s=s,
            zorder=zorder,
            facecolor=facecolor[i],
            edgecolor=edgecolor[i],
            linewidth=linewidth,
            alpha=alpha,
            marker=marker,
            **kwargs,
        )


def set_spine_tick_params(
    ax: Axes,
    spinewidth: float = 0.25,
    tickwidth: float = 0.25,
    ticklength: float = 3,
    ticklabelpad: float = 2,
    spines: Tuple[str, ...] = ("top", "right", "bottom", "left"),
) -> None:
    """
    Set spine and tick widths and lengths.

    Args:
        ax: Matplotlib axes object.
        spinewidth: Width of spines.
        tickwidth: Width of ticks.
        ticklength: Length of ticks.
        ticklabelpad: Padding between ticks and labels.
        spines: Tuple of spine names to adjust.

    Returns:
        None
    """
    for s in spines:
        ax.spines[s].set_linewidth(spinewidth)
    ax.tick_params(axis="both", width=tickwidth, length=ticklength, pad=ticklabelpad)


def scatter_on_violins_with_best(
    data: np.ndarray,
    ax: Axes,
    scatter_best: bool,
    scatter_all: bool,
    xticks: Optional[np.ndarray] = None,
    facecolor: Union[str, Tuple[float, float, float]] = "none",
    edgecolor: Union[str, Tuple[float, float, float]] = "k",
    best_scatter_alpha: float = 1.0,
    all_scatter_alpha: float = 0.35,
    best_index: Optional[int] = None,
    best_color: Optional[Union[str, Tuple[float, float, float]]] = None,
    all_marker: str = "o",
    best_marker: str = "o",
    linewidth: float = 0.5,
    best_linewidth: float = 0.75,
    uniform: List[float] = [-0.35, 0.35],
    **kwargs,
) -> None:
    """
    Scatter data points on violin plots, optionally highlighting the best point.

    Args:
        data: Array of shape (n_samples, n_variables).
        ax: Matplotlib axes object to plot on.
        scatter_best: Whether to scatter the best point.
        scatter_all: Whether to scatter all points.
        xticks: X-axis tick positions.
        facecolor: Color for marker face.
        edgecolor: Color for marker edge.
        best_scatter_alpha: Alpha for best point.
        all_scatter_alpha: Alpha for all other points.
        best_index: Index of the best point.
        best_color: Color for the best point.
        all_marker: Marker style for all points.
        best_marker: Marker style for the best point.
        linewidth: Width of marker edge for all points.
        best_linewidth: Width of marker edge for the best point.
        uniform: Range for uniform distribution of x-positions.

    Returns:
        None
    """
    if scatter_all and not scatter_best:
        scatter_on_violins_or_bars(
            data,
            ax,
            xticks=xticks,
            zorder=100,
            facecolor=facecolor,
            edgecolor=edgecolor,
            alpha=all_scatter_alpha,
            uniform=uniform,
            marker=all_marker,
            linewidth=linewidth,
        )
    elif scatter_all:
        assert (
            best_index is not None
        ), "`best_index` must be provided if `scatter_best=True`"
        indices = list(range(data.shape[0]))
        indices.remove(best_index)
        scatter_on_violins_or_bars(
            data,
            ax,
            xticks=xticks,
            indices=indices,
            zorder=10,
            facecolor=facecolor,
            edgecolor=edgecolor,
            alpha=all_scatter_alpha,
            uniform=uniform,
            marker=all_marker,
            linewidth=linewidth,
        )
    if scatter_best:
        assert (
            best_index is not None
        ), "`best_index` must be provided if `scatter_best=True`"
        assert (
            best_color is not None
        ), "`best_color` must be provided if `scatter_all=True`"
        scatter_on_violins_or_bars(
            data,
            ax,
            xticks=xticks,
            indices=[best_index],
            alpha=best_scatter_alpha,
            linewidth=best_linewidth,
            edgecolor=best_color,
            facecolor=best_color,
            uniform=[0, 0],
            s=7.5,
            zorder=11,
            marker=best_marker,
        )


def trim_axis(ax: Axes, xaxis: bool = True, yaxis: bool = True) -> None:
    """
    Trim axis to show only the range of data.

    Args:
        ax: Matplotlib axes object.
        xaxis: Whether to trim x-axis.
        yaxis: Whether to trim y-axis.

    Returns:
        None
    """
    if xaxis:
        xticks = np.array(ax.get_xticks())
        minor_xticks = np.array(ax.get_xticks(minor=True))
        all_ticks = np.sort(np.concatenate((minor_xticks, xticks)))
        if hasattr(xticks, "size"):
            firsttick = np.compress(all_ticks >= min(ax.get_xlim()), all_ticks)[0]
            lasttick = np.compress(all_ticks <= max(ax.get_xlim()), all_ticks)[-1]
            ax.spines["top"].set_bounds(firsttick, lasttick)
            ax.spines["bottom"].set_bounds(firsttick, lasttick)
            new_minor_ticks = minor_xticks.compress(minor_xticks <= lasttick)
            new_minor_ticks = new_minor_ticks.compress(new_minor_ticks >= firsttick)
            newticks = xticks.compress(xticks <= lasttick)
            newticks = newticks.compress(newticks >= firsttick)
            ax.set_xticks(newticks)
            ax.set_xticks(new_minor_ticks, minor=True)

    if yaxis:
        yticks = np.array(ax.get_yticks())
        minor_yticks = np.array(ax.get_yticks(minor=True))
        all_ticks = np.sort(np.concatenate((minor_yticks, yticks)))
        if hasattr(yticks, "size"):
            firsttick = np.compress(all_ticks >= min(ax.get_ylim()), all_ticks)[0]
            lasttick = np.compress(all_ticks <= max(ax.get_ylim()), all_ticks)[-1]
            ax.spines["left"].set_bounds(firsttick, lasttick)
            ax.spines["right"].set_bounds(firsttick, lasttick)
            new_minor_ticks = minor_yticks.compress(minor_yticks <= lasttick)
            new_minor_ticks = new_minor_ticks.compress(new_minor_ticks >= firsttick)
            newticks = yticks.compress(yticks <= lasttick)
            newticks = newticks.compress(newticks >= firsttick)
            ax.set_yticks(newticks)
            ax.set_yticks(new_minor_ticks, minor=True)


def display_significance_value(
    ax: Axes,
    pvalue: float,
    y: float,
    x0: Optional[float] = None,
    x1: Optional[float] = None,
    ticklabel: Optional[str] = None,
    bar_width: float = 0.7,
    pthresholds: Dict[float, str] = {0.01: "***", 0.05: "**", 0.1: "*"},
    fontsize: int = 8,
    annotate_insignificant: str = "",
    append_tick: bool = False,
    show_bar: bool = False,
    other_ax: Optional[Axes] = None,
    bar_height_ylim_ratio: float = 0.01,
    linewidth: float = 0.5,
    annotate_pthresholds: bool = True,
    loc_pthresh_annotation: Tuple[float, float] = (0.1, 0.1),
    location: Literal["above", "below"] = "above",
    asterisk_offset: Optional[float] = None,
) -> None:
    """
    Display a significance value annotation along x at height y.

    Args:
        ax: Matplotlib axes object.
        pvalue: P-value to display.
        y: Height to put text.
        x0: Left edge of bar if show_bar is True.
        x1: Right edge of bar if show_bar is True.
        ticklabel: Tick label to annotate.
        bar_width: Width of the bar.
        pthresholds: Dictionary of p-value thresholds and corresponding annotations.
        fontsize: Font size for annotations.
        annotate_insignificant: Annotation for insignificant p-values.
        append_tick: Whether to append annotation to tick label.
        show_bar: Whether to show a bar above the annotation.
        other_ax: Another axes object to get tick labels from.
        bar_height_ylim_ratio: Ratio of bar height to y-axis limits.
        linewidth: Line width for the bar.
        annotate_pthresholds: Whether to annotate p-value thresholds.
        loc_pthresh_annotation: Location of p-value threshold annotation.
        location: Location of annotation ("above" or "below").
        asterisk_offset: Offset for asterisk annotation.

    Returns:
        None
    """
    if x0 is None and x1 is None and ticklabel is None and bar_width is None:
        raise ValueError("specify (x0, x1) or (ticklabel, bar_width)")

    if show_bar and ((x0 is None or x1 is None) and bar_width is None):
        raise ValueError("need to specify width of bar or specify x0 and x1")

    if location == "above":
        va = "bottom"
        asterisk_offset = asterisk_offset or -0.1
    elif location == "below":
        va = "top"
        asterisk_offset = asterisk_offset or -0.05
    else:
        raise ValueError(f"location {location}")

    if x0 is None and x1 is None and ticklabel is not None:
        ticklabels = ax.get_xticklabels()
        if not ticklabels:
            ticklabels = other_ax.get_xticklabels()
        if not ticklabels:
            raise AssertionError("no ticklables found")
        tick = [tick for tick in ticklabels if tick.get_text() == ticklabel][0]
        x, _ = tick.get_position()
        x0 = x - bar_width / 2
        x1 = x + bar_width / 2

    text = ""
    any_thresh = False
    less = []
    for thresh in pthresholds:
        if pvalue < thresh:
            less.append(thresh)
            any_thresh = True

    if (any_thresh or annotate_insignificant) and show_bar:
        bar_height = (ax.get_ylim()[1] - ax.get_ylim()[0]) * bar_height_ylim_ratio
        bar_x = [x0, x0, x1, x1]
        if location == "above":
            bar_y = [y, y + bar_height, y + bar_height, y]
            y = y + bar_height
            mid = ((x0 + x1) / 2, y)
        elif location == "below":
            bar_y = [y, y - bar_height, y - bar_height, y]
            y = y - bar_height
            mid = ((x0 + x1) / 2, y)
        ax.plot(bar_x, bar_y, c="k", lw=linewidth)
        x = mid[0]

    if any_thresh:
        text = pthresholds[min(less)]
        if ticklabel is not None and append_tick:
            tick.set_text(f"{tick.get_text()}$^{{{text}}}$")
            ax.xaxis.set_ticklabels(ax.xaxis.get_ticklabels())
        else:
            ax.annotate(
                text,
                (x, y + asterisk_offset),
                fontsize=fontsize,
                ha="center",
                va=va,
            )

    elif annotate_insignificant:
        if ticklabel is not None and append_tick:
            tick.set_text(f"{tick.get_text()}$^{{{annotate_insignificant}}}$")
            ax.xaxis.set_ticklabels(ax.xaxis.get_ticklabels())
        else:
            ax.annotate(
                annotate_insignificant,
                (x, y),
                fontsize=fontsize,
                ha="center",
                va=va,
            )

    if annotate_pthresholds:
        pthreshold_annotation = ""
        for i, (thresh, symbol) in enumerate(pthresholds.items()):
            pthreshold_annotation += f"{symbol}p<{thresh:.2f}"
            if i != len(pthresholds) - 1:
                pthreshold_annotation += "\n"

        ax.annotate(
            pthreshold_annotation,
            loc_pthresh_annotation,
            xycoords="axes fraction",
            fontsize=fontsize,
            va="bottom",
            ha="left",
        )


def display_pvalues(
    ax: Axes,
    pvalues: Dict[str, float],
    ticklabels: List[str],
    data: np.ndarray,
    location: Literal["above", "below"] = "above",
    bar_width: float = 0.7,
    show_bar: bool = True,
    bar_height_ylim_ratio: float = 0.01,
    fontsize: int = 6,
    annotate_insignificant: str = "ns",
    loc_pthresh_annotation: Tuple[float, float] = (0.01, 0.01),
    append_tick: bool = False,
    data_relative_offset: float = 0.05,
    asterisk_offset: float = 0,
    pthresholds: Dict[float, str] = {0.01: "***", 0.05: "**", 0.1: "*"},
) -> None:
    """
    Annotate all p-values from a dictionary of x-tick labels to p-values.

    Args:
        ax: Matplotlib axes object.
        pvalues: Dictionary mapping x-tick labels to p-values.
        ticklabels: List of x-tick labels.
        data: Array of shape (random variables, ...).
        location: Location of annotation ("above" or "below").
        bar_width: Width of the bar.
        show_bar: Whether to show a bar above the annotation.
        bar_height_ylim_ratio: Ratio of bar height to y-axis limits.
        fontsize: Font size for annotations.
        annotate_insignificant: Annotation for insignificant p-values.
        loc_pthresh_annotation: Location of p-value threshold annotation.
        append_tick: Whether to append annotation to tick label.
        data_relative_offset: Relative offset for annotation placement.
        asterisk_offset: Offset for asterisk annotation.
        pthresholds: Dictionary of p-value thresholds and corresponding annotations.

    Returns:
        None
    """
    for key in pvalues:
        if key not in ticklabels:
            raise ValueError(f"pvalue key {key} is not a ticklabel")

    offset = data_relative_offset * np.abs(data.max() - data.min())

    ylim = ax.get_ylim()
    bars = []
    for ticklabel, pvalue in pvalues.items():
        index = [i for i, _ticklabel in enumerate(ticklabels) if _ticklabel == ticklabel][
            0
        ]
        _values = data[index]

        if location == "above":
            _max = _values.max()
            y = min(_max + offset, ylim[1])
        elif location == "below":
            _min = _values.min()
            y = max(_min - offset, ylim[0])

        display_significance_value(
            ax,
            pvalue,
            y=y,
            ticklabel=str(ticklabel),
            bar_width=bar_width,
            show_bar=show_bar,
            bar_height_ylim_ratio=bar_height_ylim_ratio,
            fontsize=fontsize,
            annotate_insignificant=annotate_insignificant,
            loc_pthresh_annotation=loc_pthresh_annotation,
            append_tick=append_tick,
            location=location,
            asterisk_offset=asterisk_offset,
            pthresholds=pthresholds,
        )
        bars.append(y)

    ax.set_ylim(*get_lims([bars, ylim], 0.01))


def closest_divisors(n: int) -> Tuple[int, int]:
    """
    Find the closest divisors of a number.

    Args:
        n: Number to find divisors for.

    Returns:
        Tuple of closest divisors.
    """
    closest_diff = float("inf")
    closest_divisors = (1, 1)

    for divisor in range(1, int(n**0.5) + 1):
        if n % divisor == 0:
            other_divisor = n // divisor
            diff = abs(divisor - other_divisor)

            if diff < closest_diff:
                closest_diff = diff
                closest_divisors = (divisor, other_divisor)

    return closest_divisors


def standalone_legend(
    labels: List[str],
    colors: List[Union[str, Tuple[float, float, float]]],
    legend_elements: Optional[List] = None,
    alpha: float = 1,
    fontsize: int = 6,
    fig: Optional[plt.Figure] = None,
    ax: Optional[Axes] = None,
    lw: float = 4,
    labelspacing: float = 0.5,
    handlelength: float = 2.0,
    n_cols: Optional[int] = None,
    columnspacing: float = 0.8,
    figsize: Optional[Tuple[float, float]] = None,
    linestyles: Optional[List[str]] = None,
) -> Tuple[plt.Figure, Axes]:
    """
    Create a standalone legend.

    Args:
        labels: List of labels for legend entries.
        colors: List of colors for legend entries.
        legend_elements: List of custom legend elements.
        alpha: Alpha value for legend entries.
        fontsize: Font size for legend text.
        fig: Existing figure to use.
        ax: Existing axes to use.
        lw: Line width for legend entries.
        labelspacing: Vertical space between legend entries.
        handlelength: Length of the legend handles.
        n_cols: Number of columns in the legend.
        columnspacing: Spacing between legend columns.
        figsize: Figure size (width, height) in inches.
        linestyles: List of line styles for legend entries.

    Returns:
        Tuple containing the figure and axes objects.
    """
    if legend_elements is None:
        from matplotlib.lines import Line2D

        legend_elements = []
        for i, label in enumerate(labels):
            legend_elements.append(
                Line2D(
                    [0],
                    [0],
                    color=colors[i],
                    lw=lw,
                    label=label,
                    alpha=alpha,
                    solid_capstyle="round",
                    linestyle=linestyles[i] if linestyles is not None else "solid",
                )
            )
    if n_cols is None:
        n_rows, n_cols = closest_divisors(
            len(legend_elements) - (len(legend_elements) % 2)
        )
    else:
        n_rows = int(np.ceil(len(labels) / n_cols))
    if fig is None or ax is None:
        figsize = figsize or [0.1 * n_cols, 0.1 * n_rows]
        fig, ax = plt.subplots(figsize=figsize)
    ax.legend(
        handles=legend_elements,
        loc="center",
        edgecolor="white",
        framealpha=1,
        fontsize=fontsize,
        labelspacing=labelspacing,
        handlelength=handlelength,
        columnspacing=columnspacing,
        ncol=n_cols,
    )
    rm_spines(ax, rm_yticks=True, rm_xticks=True)
    return fig, ax


def extend_arg(
    arg: Union[Number, List[Number]],
    argtype: type,
    r: np.ndarray,
    default: Any,
    dim: int = -1,
) -> Union[List[Number], Number]:
    """
    Extend an argument to the correct length for a given dimension.

    Args:
        arg: Argument to extend.
        argtype: Type of the argument.
        r: Reference array for shape.
        default: Default value if arg is not provided.
        dim: Dimension to extend along.

    Returns:
        Extended argument.
    """
    r = np.asarray(r)

    if isinstance(arg, argtype) and r.ndim > 1:
        return [arg] * r.shape[dim]
    elif (
        isinstance(arg, Iterable)
        and len(arg) == r.shape[dim]
        or r.ndim == 1
        and np.asarray(arg).size == 1
    ):
        return arg
    elif r.ndim == 1:
        return default
    else:
        raise ValueError(
            f"arg must be either an integer or a list of length {r.shape[-1]}."
        )


# colormap from
# https://stackoverflow.com/questions/23712207/cyclic-colormap-without-visual-distortions-for-use-in-phase-angle-plots
cm_uniform_2d = np.array([
    [0.91510904, 0.55114749, 0.67037311],
    [0.91696411, 0.55081563, 0.66264366],
    [0.91870995, 0.55055664, 0.65485881],
    [0.92034498, 0.55037149, 0.64702356],
    [0.92186763, 0.55026107, 0.63914306],
    [0.92327636, 0.55022625, 0.63122259],
    [0.9245696, 0.55026781, 0.62326754],
    [0.92574582, 0.5503865, 0.6152834],
    [0.92680349, 0.55058299, 0.6072758],
    [0.92774112, 0.55085789, 0.59925045],
    [0.9285572, 0.55121174, 0.59121319],
    [0.92925027, 0.551645, 0.58316992],
    [0.92981889, 0.55215808, 0.57512667],
    [0.93026165, 0.55275127, 0.56708953],
    [0.93057716, 0.5534248, 0.55906469],
    [0.93076407, 0.55417883, 0.55105838],
    [0.93082107, 0.55501339, 0.54307696],
    [0.93074689, 0.55592845, 0.53512681],
    [0.9305403, 0.55692387, 0.52721438],
    [0.93020012, 0.55799943, 0.51934621],
    [0.92972523, 0.55915477, 0.51152885],
    [0.92911454, 0.56038948, 0.50376893],
    [0.92836703, 0.56170301, 0.49607312],
    [0.92748175, 0.56309471, 0.48844813],
    [0.9264578, 0.56456383, 0.48090073],
    [0.92529434, 0.56610951, 0.47343769],
    [0.92399062, 0.56773078, 0.46606586],
    [0.92254595, 0.56942656, 0.45879209],
    [0.92095971, 0.57119566, 0.4516233],
    [0.91923137, 0.5730368, 0.44456642],
    [0.91736048, 0.57494856, 0.4376284],
    [0.91534665, 0.57692945, 0.43081625],
    [0.91318962, 0.57897785, 0.42413698],
    [0.91088917, 0.58109205, 0.41759765],
    [0.90844521, 0.58327024, 0.41120533],
    [0.90585771, 0.58551053, 0.40496711],
    [0.90312676, 0.5878109, 0.3988901],
    [0.90025252, 0.59016928, 0.39298143],
    [0.89723527, 0.5925835, 0.38724821],
    [0.89407538, 0.59505131, 0.38169756],
    [0.89077331, 0.59757038, 0.37633658],
    [0.88732963, 0.60013832, 0.37117234],
    [0.88374501, 0.60275266, 0.36621186],
    [0.88002022, 0.6054109, 0.36146209],
    [0.87615612, 0.60811044, 0.35692989],
    [0.87215369, 0.61084868, 0.352622],
    [0.86801401, 0.61362295, 0.34854502],
    [0.86373824, 0.61643054, 0.34470535],
    [0.85932766, 0.61926872, 0.3411092],
    [0.85478365, 0.62213474, 0.3377625],
    [0.85010767, 0.6250258, 0.33467091],
    [0.84530131, 0.62793914, 0.3318397],
    [0.84036623, 0.63087193, 0.32927381],
    [0.8353042, 0.63382139, 0.32697771],
    [0.83011708, 0.63678472, 0.32495541],
    [0.82480682, 0.63975913, 0.32321038],
    [0.81937548, 0.64274185, 0.32174556],
    [0.81382519, 0.64573011, 0.32056327],
    [0.80815818, 0.6487212, 0.31966522],
    [0.80237677, 0.65171241, 0.31905244],
    [0.79648336, 0.65470106, 0.31872531],
    [0.79048044, 0.65768455, 0.31868352],
    [0.78437059, 0.66066026, 0.31892606],
    [0.77815645, 0.66362567, 0.31945124],
    [0.77184076, 0.66657827, 0.32025669],
    [0.76542634, 0.66951562, 0.3213394],
    [0.75891609, 0.67243534, 0.32269572],
    [0.75231298, 0.67533509, 0.32432138],
    [0.74562004, 0.6782126, 0.32621159],
    [0.73884042, 0.68106567, 0.32836102],
    [0.73197731, 0.68389214, 0.33076388],
    [0.72503398, 0.68668995, 0.33341395],
    [0.7180138, 0.68945708, 0.33630465],
    [0.71092018, 0.69219158, 0.33942908],
    [0.70375663, 0.69489159, 0.34278007],
    [0.69652673, 0.69755529, 0.34635023],
    [0.68923414, 0.70018097, 0.35013201],
    [0.6818826, 0.70276695, 0.35411772],
    [0.67447591, 0.70531165, 0.3582996],
    [0.667018, 0.70781354, 0.36266984],
    [0.65951284, 0.71027119, 0.36722061],
    [0.65196451, 0.71268322, 0.37194411],
    [0.64437719, 0.71504832, 0.37683259],
    [0.63675512, 0.71736525, 0.38187838],
    [0.62910269, 0.71963286, 0.38707389],
    [0.62142435, 0.72185004, 0.39241165],
    [0.61372469, 0.72401576, 0.39788432],
    [0.60600841, 0.72612907, 0.40348469],
    [0.59828032, 0.72818906, 0.40920573],
    [0.59054536, 0.73019489, 0.41504052],
    [0.58280863, 0.73214581, 0.42098233],
    [0.57507535, 0.7340411, 0.42702461],
    [0.5673509, 0.7358801, 0.43316094],
    [0.55964082, 0.73766224, 0.43938511],
    [0.55195081, 0.73938697, 0.44569104],
    [0.54428677, 0.74105381, 0.45207286],
    [0.53665478, 0.74266235, 0.45852483],
    [0.52906111, 0.74421221, 0.4650414],
    [0.52151225, 0.74570306, 0.47161718],
    [0.5140149, 0.74713464, 0.47824691],
    [0.506576, 0.74850672, 0.48492552],
    [0.49920271, 0.74981912, 0.49164808],
    [0.49190247, 0.75107171, 0.4984098],
    [0.48468293, 0.75226438, 0.50520604],
    [0.47755205, 0.7533971, 0.51203229],
    [0.47051802, 0.75446984, 0.5188842],
    [0.46358932, 0.75548263, 0.52575752],
    [0.45677469, 0.75643553, 0.53264815],
    [0.45008317, 0.75732863, 0.5395521],
    [0.44352403, 0.75816207, 0.54646551],
    [0.43710682, 0.758936, 0.55338462],
    [0.43084133, 0.7596506, 0.56030581],
    [0.42473758, 0.76030611, 0.56722555],
    [0.41880579, 0.76090275, 0.5741404],
    [0.41305637, 0.76144081, 0.58104704],
    [0.40749984, 0.76192057, 0.58794226],
    [0.40214685, 0.76234235, 0.59482292],
    [0.39700806, 0.7627065, 0.60168598],
    [0.39209414, 0.76301337, 0.6085285],
    [0.38741566, 0.76326334, 0.6153476],
    [0.38298304, 0.76345681, 0.62214052],
    [0.37880647, 0.7635942, 0.62890454],
    [0.37489579, 0.76367593, 0.63563704],
    [0.37126045, 0.76370246, 0.64233547],
    [0.36790936, 0.76367425, 0.64899736],
    [0.36485083, 0.76359176, 0.6556203],
    [0.36209245, 0.76345549, 0.66220193],
    [0.359641, 0.76326594, 0.66873999],
    [0.35750235, 0.76302361, 0.67523226],
    [0.35568141, 0.76272903, 0.68167659],
    [0.35418202, 0.76238272, 0.68807086],
    [0.3530069, 0.76198523, 0.69441305],
    [0.35215761, 0.7615371, 0.70070115],
    [0.35163454, 0.76103888, 0.70693324],
    [0.35143685, 0.76049114, 0.71310742],
    [0.35156253, 0.75989444, 0.71922184],
    [0.35200839, 0.75924936, 0.72527472],
    [0.3527701, 0.75855647, 0.73126429],
    [0.3538423, 0.75781637, 0.73718884],
    [0.3552186, 0.75702964, 0.7430467],
    [0.35689171, 0.75619688, 0.74883624],
    [0.35885353, 0.75531868, 0.75455584],
    [0.36109522, 0.75439565, 0.76020396],
    [0.36360734, 0.75342839, 0.76577905],
    [0.36637995, 0.75241752, 0.77127961],
    [0.3694027, 0.75136364, 0.77670417],
    [0.37266493, 0.75026738, 0.7820513],
    [0.37615579, 0.74912934, 0.78731957],
    [0.37986429, 0.74795017, 0.79250759],
    [0.38377944, 0.74673047, 0.797614],
    [0.38789026, 0.74547088, 0.80263746],
    [0.3921859, 0.74417203, 0.80757663],
    [0.39665568, 0.74283455, 0.81243022],
    [0.40128912, 0.74145908, 0.81719695],
    [0.406076, 0.74004626, 0.82187554],
    [0.41100641, 0.73859673, 0.82646476],
    [0.41607073, 0.73711114, 0.83096336],
    [0.4212597, 0.73559013, 0.83537014],
    [0.42656439, 0.73403435, 0.83968388],
    [0.43197625, 0.73244447, 0.8439034],
    [0.43748708, 0.73082114, 0.84802751],
    [0.44308905, 0.72916502, 0.85205505],
    [0.44877471, 0.72747678, 0.85598486],
    [0.45453694, 0.72575709, 0.85981579],
    [0.46036897, 0.72400662, 0.8635467],
    [0.4662644, 0.72222606, 0.86717646],
    [0.47221713, 0.72041608, 0.87070395],
    [0.47822138, 0.71857738, 0.87412804],
    [0.4842717, 0.71671065, 0.87744763],
    [0.4903629, 0.71481659, 0.88066162],
    [0.49649009, 0.71289591, 0.8837689],
    [0.50264864, 0.71094931, 0.88676838],
    [0.50883417, 0.70897752, 0.88965898],
    [0.51504253, 0.70698127, 0.89243961],
    [0.52126981, 0.70496128, 0.8951092],
    [0.52751231, 0.70291829, 0.89766666],
    [0.53376652, 0.70085306, 0.90011093],
    [0.54002912, 0.69876633, 0.90244095],
    [0.54629699, 0.69665888, 0.90465565],
    [0.55256715, 0.69453147, 0.90675397],
    [0.55883679, 0.69238489, 0.90873487],
    [0.56510323, 0.69021993, 0.9105973],
    [0.57136396, 0.68803739, 0.91234022],
    [0.57761655, 0.68583808, 0.91396258],
    [0.58385872, 0.68362282, 0.91546336],
    [0.59008831, 0.68139246, 0.91684154],
    [0.59630323, 0.67914782, 0.9180961],
    [0.60250152, 0.67688977, 0.91922603],
    [0.60868128, 0.67461918, 0.92023033],
    [0.61484071, 0.67233692, 0.921108],
    [0.62097809, 0.67004388, 0.92185807],
    [0.62709176, 0.66774097, 0.92247957],
    [0.63318012, 0.66542911, 0.92297153],
    [0.63924166, 0.66310923, 0.92333301],
    [0.64527488, 0.66078227, 0.92356308],
    [0.65127837, 0.65844919, 0.92366082],
    [0.65725076, 0.65611096, 0.92362532],
    [0.66319071, 0.65376857, 0.92345572],
    [0.66909691, 0.65142302, 0.92315115],
    [0.67496813, 0.64907533, 0.92271076],
    [0.68080311, 0.64672651, 0.92213374],
    [0.68660068, 0.64437763, 0.92141929],
    [0.69235965, 0.64202973, 0.92056665],
    [0.69807888, 0.6396839, 0.91957507],
    [0.70375724, 0.63734122, 0.91844386],
    [0.70939361, 0.63500279, 0.91717232],
    [0.7149869, 0.63266974, 0.91575983],
    [0.72053602, 0.63034321, 0.91420578],
    [0.72603991, 0.62802433, 0.9125096],
    [0.7314975, 0.62571429, 0.91067077],
    [0.73690773, 0.62341425, 0.9086888],
    [0.74226956, 0.62112542, 0.90656328],
    [0.74758193, 0.61884899, 0.90429382],
    [0.75284381, 0.6165862, 0.90188009],
    [0.75805413, 0.61433829, 0.89932181],
    [0.76321187, 0.6121065, 0.89661877],
    [0.76831596, 0.6098921, 0.89377082],
    [0.77336536, 0.60769637, 0.89077786],
    [0.77835901, 0.6055206, 0.88763988],
    [0.78329583, 0.6033661, 0.88435693],
    [0.78817477, 0.60123418, 0.88092913],
    [0.79299473, 0.59912616, 0.87735668],
    [0.79775462, 0.59704339, 0.87363986],
    [0.80245335, 0.59498722, 0.86977904],
    [0.8070898, 0.592959, 0.86577468],
    [0.81166284, 0.5909601, 0.86162732],
    [0.81617134, 0.5889919, 0.8573376],
    [0.82061414, 0.58705579, 0.85290625],
    [0.82499007, 0.58515315, 0.84833413],
    [0.82929796, 0.58328538, 0.84362217],
    [0.83353661, 0.58145389, 0.83877142],
    [0.8377048, 0.57966009, 0.83378306],
    [0.8418013, 0.57790538, 0.82865836],
    [0.84582486, 0.57619119, 0.82339871],
    [0.84977422, 0.57451892, 0.81800565],
    [0.85364809, 0.57289, 0.8124808],
    [0.85744519, 0.57130585, 0.80682595],
    [0.86116418, 0.56976788, 0.80104298],
    [0.86480373, 0.56827749, 0.79513394],
    [0.86836249, 0.56683612, 0.789101],
    [0.87183909, 0.56544515, 0.78294645],
    [0.87523214, 0.56410599, 0.77667274],
    [0.87854024, 0.56282002, 0.77028247],
    [0.88176195, 0.56158863, 0.76377835],
    [0.88489584, 0.56041319, 0.75716326],
    [0.88794045, 0.55929505, 0.75044023],
    [0.89089432, 0.55823556, 0.74361241],
    [0.89375596, 0.55723605, 0.73668312],
    [0.89652387, 0.55629781, 0.72965583],
    [0.89919653, 0.55542215, 0.72253414],
    [0.90177242, 0.55461033, 0.71532181],
    [0.90425, 0.55386358, 0.70802274],
    [0.90662774, 0.55318313, 0.70064098],
    [0.90890408, 0.55257016, 0.69318073],
    [0.91107745, 0.55202582, 0.68564633],
    [0.91314629, 0.55155124, 0.67804225],
])

cm_uniform_2d = colors.ListedColormap(cm_uniform_2d)
try:
    cm.get_cmap("cm_uniform_2d")
except ValueError:
    cm.register(cm_uniform_2d, name="cm_uniform_2d")



"""Functions to calculate figure sizes for plotting."""

from dataclasses import dataclass
from typing import Tuple, Union

from . import plt_utils


def figsize_from_n_items(
    n_panels: int,
    max_figure_height_cm: float = 22,
    panel_height_cm: float = 3,
    max_figure_width_cm: float = 18,
    panel_width_cm: float = 3.6,
    dw_cm: float = 0.1,
) -> "FigsizeCM":
    """
    Calculate figure size based on the number of panels.

    Args:
        n_panels: Number of panels in the figure.
        max_figure_height_cm: Maximum figure height in centimeters.
        panel_height_cm: Height of each panel in centimeters.
        max_figure_width_cm: Maximum figure width in centimeters.
        panel_width_cm: Width of each panel in centimeters.
        dw_cm: Decrement width in centimeters for panel size adjustment.

    Returns:
        FigsizeCM: Calculated figure size.
    """
    n_columns = int(max_figure_width_cm / panel_width_cm)
    n_rows = 1
    while n_columns * n_rows < n_panels:
        n_rows += 1
    return fit_panel_size(
        n_rows,
        n_columns,
        max_figure_height_cm,
        panel_height_cm,
        max_figure_width_cm,
        panel_width_cm,
        dw_cm,
    )


@dataclass
class FigsizeCM:
    """
    Represents figure size in centimeters.

    Attributes:
        n_rows: Number of rows in the figure.
        n_columns: Number of columns in the figure.
        height: Height of the figure in centimeters.
        width: Width of the figure in centimeters.
        pad: Padding in centimeters.
    """

    n_rows: int
    n_columns: int
    height: float
    width: float
    pad: float = 0.5

    @property
    def inches_wh(self) -> Tuple[float, float]:
        """Convert width and height to inches."""
        return cm_to_inch(self.width + self.pad, self.height + self.pad)

    @property
    def panel_height_cm(self) -> float:
        """Calculate panel height in centimeters."""
        return self.height / self.n_rows

    @property
    def panel_width_cm(self) -> float:
        """Calculate panel width in centimeters."""
        return self.width / self.n_columns

    def axis_grid(
        self,
        projection: Union[str, None] = None,
        as_matrix: bool = False,
        fontsize: int = 5,
        wspace: float = 0.1,
        hspace: float = 0.3,
        alpha: float = 1,
        unmask_n: Union[int, None] = None,
    ) -> Tuple:
        """
        Create an axis grid for the figure.

        Args:
            projection: Type of projection for the axes.
            as_matrix: Whether to return axes as a matrix.
            fontsize: Font size for the axes.
            wspace: Width space between subplots.
            hspace: Height space between subplots.
            alpha: Alpha value for the axes.
            unmask_n: Number of axes to unmask.

        Returns:
            Tuple containing the figure and axes.
        """
        fig, axes, _ = plt_utils.get_axis_grid(
            gridwidth=self.n_columns,
            gridheight=self.n_rows,
            figsize=self.inches_wh,
            projection=projection,
            as_matrix=as_matrix,
            fontsize=fontsize,
            wspace=wspace,
            hspace=hspace,
            alpha=alpha,
            unmask_n=unmask_n,
        )
        return fig, axes


def figure_size_cm(
    n_panel_rows: int,
    n_panel_columns: int,
    max_figure_height_cm: float = 22,
    panel_height_cm: float = 3,
    max_figure_width_cm: float = 18,
    panel_width_cm: float = 3.6,
    allow_rearranging: bool = True,
) -> FigsizeCM:
    """
    Calculate figure size in centimeters.

    Args:
        n_panel_rows: Number of panel rows.
        n_panel_columns: Number of panel columns.
        max_figure_height_cm: Maximum figure height in centimeters.
        panel_height_cm: Height of each panel in centimeters.
        max_figure_width_cm: Maximum figure width in centimeters.
        panel_width_cm: Width of each panel in centimeters.
        allow_rearranging: Whether to allow rearranging panels.

    Returns:
        FigsizeCM: Calculated figure size.

    Raises:
        ValueError: If the figure size is not realizable under given constraints.
    """
    width = n_panel_columns * panel_width_cm
    height = n_panel_rows * panel_height_cm
    n_panels = n_panel_rows * n_panel_columns

    if width > max_figure_width_cm and height > max_figure_height_cm:
        raise ValueError("Not realizable under given size constraints")
    elif width > max_figure_width_cm and allow_rearranging:
        n_panel_columns -= 1
        while n_panel_columns * n_panel_rows < n_panels:
            n_panel_rows += 1
        return figure_size_cm(
            n_panel_rows=n_panel_rows,
            n_panel_columns=n_panel_columns,
            max_figure_height_cm=max_figure_height_cm,
            panel_height_cm=panel_height_cm,
            max_figure_width_cm=max_figure_width_cm,
            panel_width_cm=panel_width_cm,
        )
    elif height > max_figure_height_cm and allow_rearranging:
        n_panel_rows -= 1
        while n_panel_columns * n_panel_rows < n_panels:
            n_panel_columns += 1
        return figure_size_cm(
            n_panel_rows=n_panel_rows,
            n_panel_columns=n_panel_columns,
            max_figure_height_cm=max_figure_height_cm,
            panel_height_cm=panel_height_cm,
            max_figure_width_cm=max_figure_width_cm,
            panel_width_cm=panel_width_cm,
        )
    elif not allow_rearranging and (
        width > max_figure_width_cm or height > max_figure_height_cm
    ):
        raise ValueError("Not realizable under given size constraints")

    return FigsizeCM(n_panel_rows, n_panel_columns, height, width)


def fit_panel_size(
    n_panel_rows: int,
    n_panel_columns: int,
    max_figure_height_cm: float = 22,
    panel_height_cm: float = 3,
    max_figure_width_cm: float = 18,
    panel_width_cm: float = 3.6,
    dw_cm: float = 0.1,
    allow_rearranging: bool = True,
) -> FigsizeCM:
    """
    Fit panel size to figure constraints.

    Args:
        n_panel_rows: Number of panel rows.
        n_panel_columns: Number of panel columns.
        max_figure_height_cm: Maximum figure height in centimeters.
        panel_height_cm: Height of each panel in centimeters.
        max_figure_width_cm: Maximum figure width in centimeters.
        panel_width_cm: Width of each panel in centimeters.
        dw_cm: Decrement width in centimeters for panel size adjustment.
        allow_rearranging: Whether to allow rearranging panels.

    Returns:
        FigsizeCM: Fitted figure size.
    """
    ratio = panel_width_cm / panel_height_cm

    try:
        return figure_size_cm(
            n_panel_rows,
            n_panel_columns,
            max_figure_height_cm,
            panel_height_cm,
            max_figure_width_cm,
            panel_width_cm,
            allow_rearranging=allow_rearranging,
        )
    except ValueError:
        new_panel_width_cm = panel_width_cm - dw_cm
        new_panel_height_cm = new_panel_width_cm / ratio
        return fit_panel_size(
            n_panel_rows,
            n_panel_columns,
            max_figure_height_cm,
            new_panel_height_cm,
            max_figure_width_cm,
            new_panel_width_cm,
            dw_cm,
            allow_rearranging=allow_rearranging,
        )


def cm_to_inch(*args: Union[Tuple[float, float], float]) -> Tuple[float, float]:
    """
    Convert centimeters to inches.

    Args:
        *args: Either a tuple of (width, height) or separate width and height values.

    Returns:
        Tuple of width and height in inches.
    """
    if len(args) == 1:
        width, height = args[0]
    elif len(args) == 2:
        width, height = args
    else:
        raise ValueError("Invalid number of arguments")
    return width / 2.54, height / 2.54



"""Script to train an ensemble of models."""

import argparse
import logging
from importlib import resources
from typing import List

from flyvis.utils.compute_cloud_utils import launch_range
from flyvis.utils.config_utils import HybridArgumentParser

logging.basicConfig(
    format="[%(asctime)s] [%(filename)s:%(lineno)d] %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


def train_models(args: argparse.Namespace, kwargs: List[str]) -> None:
    """
    Launch training jobs for an ensemble of models.

    Args:
        args: Command-line arguments.
        kwargs: Additional keyword arguments as a list of strings.
    """
    launch_range(
        args.start,
        args.end,
        args.ensemble_id,
        args.task_name,
        args.nP,
        args.gpu,
        args.q,
        args.train_script,
        args.dry,
        kwargs,
    )


if __name__ == "__main__":
    CONFIG_PATH = str(resources.files("flyvis") / "config" / "solver.yaml")

    parser = HybridArgumentParser(
        description=(
            "Train an ensemble of models. Launches a job for each model on "
            "the compute cloud."
        ),
        drop_disjoint_from=CONFIG_PATH,
        formatter_class=argparse.RawTextHelpFormatter,
        usage=(
            "\nflyvis train [-h] [--start START] [--end END] [...] "
            "--ensemble_id ENSEMBLE_ID --task_name TASK_NAME "
            "[hydra_options...]\n"
            "       or\n"
            "%(prog)s [-h] [--start START] [--end END] [...] "
            "--ensemble_id ENSEMBLE_ID --task_name TASK_NAME "
            "[hydra_options...]\n"
            "\n"
            "For a full list of hydra options and default arguments, run: "
            "flyvis train-single --help"
        ),
    )
    parser.add_argument("--start", type=int, default=0, help="Start id of ensemble.")
    parser.add_argument("--end", type=int, default=50, help="End id of ensemble.")
    parser.add_argument("--nP", type=int, default=4, help="Number of processors.")
    parser.add_argument("--gpu", type=str, default="num=1", help="Number of GPUs.")
    parser.add_argument("--q", type=str, default="gpu_l4", help="Queue.")
    parser.add_argument(
        "--ensemble_id",
        type=int,
        required=True,
        help="Id of the ensemble, e.g. 0045.",
    )
    parser.add_argument(
        "--task_name",
        type=str,
        required=True,
        help="Name given to the task, e.g., flow.",
    )
    DEFAULT_SCRIPT = str(resources.files("flyvis_cli") / "training" / "train_single.py")
    parser.add_argument(
        "--train_script",
        type=str,
        default=DEFAULT_SCRIPT,
        help=f"Script to run for training. Default: {DEFAULT_SCRIPT}",
    )
    parser.add_argument(
        "--dry",
        action="store_true",
        help="Perform a dry run without actually launching jobs.",
    )

    args, _ = parser.parse_known_intermixed_args()
    kwargs = parser.hydra_argv()
    train_models(args, kwargs)



from numbers import Number
from typing import Any, Mapping

import numpy as np


def byte_to_str(obj: Any) -> Any:
    """Cast byte elements to string types recursively.

    This function recursively converts byte elements to string types in nested
    data structures.

    Args:
        obj: The object to be processed. Can be of various types including
            Mapping, numpy.ndarray, list, tuple, bytes, str, or Number.

    Returns:
        The input object with all byte elements converted to strings.

    Raises:
        TypeError: If the input object cannot be cast to a string type.

    Note:
        This function will cast all byte elements in nested lists or tuples.

    Examples:
        ```python
        >>> byte_to_str(b"hello")
        'hello'
        >>> byte_to_str([b"world", 42, {b"key": b"value"}])
        ['world', 42, {'key': 'value'}]
        ```
    """
    if isinstance(obj, Mapping):
        return type(obj)({k: byte_to_str(v) for k, v in obj.items()})
    elif isinstance(obj, np.ndarray):
        if np.issubdtype(obj.dtype, np.dtype("S")):
            return obj.astype("U")
        return obj
    elif isinstance(obj, list):
        return [byte_to_str(item) for item in obj]
    elif isinstance(obj, tuple):
        return tuple(byte_to_str(item) for item in obj)
    elif isinstance(obj, bytes):
        return obj.decode()
    elif isinstance(obj, (str, Number)):
        return obj
    else:
        raise TypeError(f"can't cast {obj} of type {type(obj)} to str")



"""Stimulus response functions with caching using xarray."""

from __future__ import annotations

import pprint
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union

import numpy as np

import flyvis
from flyvis.connectome import ReceptiveFields
from flyvis.datasets.moving_bar import MovingEdge
from flyvis.utils.activity_utils import LayerActivity, SourceCurrentView

__all__ = ["compute_currents", "generic_currents", "moving_edge_currents"]


@dataclass
class TargetData:
    activity_central: List[Any] = field(default_factory=list)
    source_data: Dict[str, List[Any]] = field(default_factory=dict)

    def __repr__(self):
        sd = {ct: np.shape(d) for ct, d in self.source_data.items()}
        formatted_sd = pprint.pformat(sd, indent=4)
        return (
            f"TargetData(activity_central={np.shape(self.activity_central)},\n"
            f"            source_data={formatted_sd})"
        )


@dataclass
class ExperimentData:
    config: Any
    checkpoint: str = ""
    target_data: Dict[str, TargetData] = field(default_factory=dict)


def compute_currents(
    network: "flyvis.network.CheckpointedNetwork",
    dataset_class: type,
    dataset_config: Dict,
    target_cell_types: Optional[List[str]] = None,
    t_pre: float = 2.0,
    t_fade_in: float = 0.0,
    dt: float = 1 / 200,
) -> ExperimentData:
    # Reconstruct the network
    network.recover()
    checkpoint = network.checkpoint
    network = network.network  # type: flyvis.network.Network

    # Initialize dataset
    dataset = dataset_class(**dataset_config)

    # Initialize the experiment data
    experiment_data = ExperimentData(config=dataset.config, checkpoint=checkpoint)

    edges = network.connectome.edges.to_df()
    target_types = target_cell_types or edges.target_type.unique()

    # to store the responses and currents in a structured way
    activity_indexer = LayerActivity(None, network.connectome, keepref=True)
    source_current_indexer = {
        target_type: SourceCurrentView(ReceptiveFields(target_type, edges), None)
        for target_type in target_types
    }

    # Initialize target_data in experiment_data
    for target_type in target_types:
        experiment_data.target_data[target_type] = TargetData()

    for _, activity, current in network.current_response(
        dataset,
        dt,
        indices=None,
        t_pre=t_pre,
        t_fade_in=t_fade_in,
    ):
        # implementing computation and writing here together to save some runtime memory

        # Update activity indexer
        activity_indexer.update(activity)

        for target_type in target_types:
            target_data = experiment_data.target_data[target_type]

            # Append central activity data
            target_data.activity_central.append(activity_indexer.central[target_type])

            # Update source current indexer
            source_current_indexer[target_type].update(current)
            for source_type in source_current_indexer[target_type].source_types:
                if source_type not in target_data.source_data:
                    target_data.source_data[source_type] = []
                # Append source current data
                target_data.source_data[source_type].append(
                    source_current_indexer[target_type][source_type]
                )

    return experiment_data


def generic_currents(
    network_view_or_ensemble: Union["flyvis.NetworkView", "flyvis.ensemble.Ensemble"],
    dataset,
    dataset_config: Dict,
    default_dataset_cls: type,
    target_cell_types: Optional[List[str]],
    t_pre: float,
    t_fade_in: float,
    dt: float,
) -> List[ExperimentData]:
    """Return responses for a given dataset as an xarray Dataset."""
    # Handle both single and multiple NetworkViews
    if isinstance(network_view_or_ensemble, flyvis.NetworkView):
        network_views = [network_view_or_ensemble]
    else:
        network_views = list(network_view_or_ensemble.values())

    # Prepare dataset class
    dataset_class = default_dataset_cls if dataset is None else type(dataset)
    dataset_config = dataset_config if dataset is None else dataset.config.to_dict()

    # quick bugfix from datasets that have type in their config but don't expect it
    dataset_config.pop("type", None)

    # Prepare list to collect datasets
    results = []
    # checkpoints = []

    def handle_network(idx, network_view, network=None):
        # Pass initialized network over to next network view to avoid
        # reinitializing the network
        checkpointed_network = network_view.network(
            checkpoint="best", network=network, lazy=True
        )

        # use the cache from this network_view
        cached_compute_responses_fn = network_view.memory.cache(
            compute_currents,
        )
        call_in_cache = cached_compute_responses_fn.check_call_in_cache(
            checkpointed_network,
            dataset_class,
            dataset_config,
            target_cell_types,
            t_pre,
            t_fade_in,
            dt,
        )

        if call_in_cache:
            # don't initialize the network when the call is in cache
            # print('call in cache')
            pass
        elif network is None and checkpointed_network.network is None:
            # initialize the network when the call is not in cache
            # and the network is not passed from the previous network view
            # print('call not in cache, init network')
            checkpointed_network.init()

        # Call the cached compute_responses function
        results.append(
            cached_compute_responses_fn(
                checkpointed_network,
                dataset_class,
                dataset_config,
                target_cell_types,
                t_pre,
                t_fade_in,
            )  # type: ExperimentData
        )
        # checkpoints.append(checkpointed_network.checkpoint)
        return checkpointed_network.network

    network = handle_network(0, network_views[0], None)

    for idx, network_view in enumerate(network_views[1:], 1):
        network = handle_network(idx, network_view, network)

    return results


def moving_edge_currents(
    network_view_or_ensemble: Union["flyvis.NetworkView", "flyvis.ensemble.Ensemble"],
    target_cell_types: Optional[List[str]] = [
        "T4a",
        "T4b",
        "T4c",
        "T4d",
        "T5a",
        "T5b",
        "T5c",
        "T5d",
        "TmY3",
    ],
    dataset: Optional[MovingEdge] = None,
    speeds=(19,),
    offsets=(-10, 11),
    angles=(0, 45, 90, 180, 225, 270),
    dt=1 / 200,
) -> List[ExperimentData]:
    default_dataset_config = dict(
        widths=[80],
        offsets=offsets,
        intensities=[0, 1],
        speeds=speeds,
        height=80,
        bar_loc_horizontal=0.0,
        shuffle_offsets=False,
        post_pad_mode="continue",
        t_pre=1.0,
        t_post=1.0,
        dt=dt,
        angles=angles,
    )
    return generic_currents(
        network_view_or_ensemble,
        dataset,
        default_dataset_config,
        MovingEdge,
        target_cell_types,
        t_pre=1.0,
        t_fade_in=0.0,
        dt=dt,
    )


if __name__ == "__main__":
    from flyvis import NetworkView

    nv = NetworkView("flow/0000/000")
    moving_edge_currents(nv)
